from typing import TYPE_CHECKING, Callable, List, Optional, Union

import audobject
from torchvision import transforms as T

from autrainer.core.structs import AbstractDataItem

from .abstract_transform import AbstractTransform


if TYPE_CHECKING:  # pragma: no cover
    from autrainer.datasets import AbstractDataset


class SmartCompose(T.Compose, audobject.Object):
    def __init__(self, transforms: List[AbstractTransform], **kwargs) -> None:
        """SmartCompose wrapper for torchvision.transforms.Compose, which
        allows for simple composition of transforms by adding them together.
        Transforms are automatically sorted by their order attribute if present.

        Additionally, SmartCompose allows for the specification of a
        collate function to be used in a DataLoader, the last collate function
        specified will be used.

        Args:
            transforms: List of transforms to compose.
            **kwargs: Additional keyword arguments to pass to
                torchvision.transforms.Compose.
        """
        super().__init__(transforms, **kwargs)
        self._sort()

    def __add__(
        self,
        other: Optional[
            Union["SmartCompose", AbstractTransform, List[AbstractTransform]]
        ],
    ) -> "SmartCompose":
        """Add another transform to the composition. Transforms are
        automatically sorted by their order attribute if present.

        Args:
            other: Transform to add to the composition.

        Raises:
            TypeError: If the addition is not a valid type.
                Supported types are AbstractTransform, SmartCompose,
                and list of AbstractTransform or SmartCompose.

        Returns:
            New SmartCompose object with the added transform.
        """
        if other is None:
            return self
        if isinstance(other, AbstractTransform):
            return SmartCompose(self.transforms + [other])
        if isinstance(other, SmartCompose):
            return SmartCompose(self.transforms + other.transforms)
        if isinstance(other, list):
            t = SmartCompose([])
            for o in other:
                if isinstance(o, AbstractTransform):
                    t.transforms.append(o)
                elif isinstance(o, SmartCompose):
                    t.transforms += o.transforms
            if len(t.transforms):
                return SmartCompose(self.transforms + t.transforms)
        raise TypeError(
            f"Unsupported type for addition: {type(other)}. "
            "Supported types are AbstractTransform, SmartCompose, "
            "and list of AbstractTransform or SmartCompose."
        )

    def _sort(self) -> None:
        """Sort the transforms by their order attribute if present."""
        self.transforms.sort(key=lambda x: getattr(x, "order", 0))

    def get_collate_fn(self, data: "AbstractDataset") -> Callable:
        """Get the collate function. If no collate function is present in
        the transforms, the dataset default is returned.
        If multiple collate functions are present, the last one is used.

        Args:
            data: Dataset to get the collate function for.
                Includes a ``default_collate_fn``.

        Returns:
            Collate function.
        """
        default_fn = data.default_collate_fn
        collate_fn = None
        for t in self.transforms:
            if fn := getattr(t, "get_collate_fn", None):
                collate_fn = fn
        if collate_fn is not None:
            return collate_fn(data, default_fn)
        return default_fn

    def setup(self, data: "AbstractDataset") -> "SmartCompose":
        for t in self.transforms:
            if hasattr(t, "setup"):
                t.setup(data)
        return self

    def offset_generator_seed(self, offset: int) -> None:
        """Offset the generator seed for transforms that use random
        number generators. Useful for ensuring reproducibility and
        randomness of augmentations when using multiple workers.

        Args:
            offset: Offset to add to the generator seed. Usually the
                worker index.
        """
        for t in self.transforms:
            if hasattr(t, "offset_generator_seed"):
                t.offset_generator_seed(offset)

    def __call__(self, item: AbstractDataItem) -> AbstractDataItem:
        """Apply the transform to the input data item.

        Args:
            item: The input data item to transform.

        Returns:
            The transformed data item.
        """
        for t in self.transforms:
            item = t(item)
        return item



import argparse
import logging
import os
from pathlib import Path
import shutil
import sys
import tempfile
from typing import Any, Optional

from hydra.experimental.callback import Callback
from omegaconf import DictConfig


def pop_config_dir_from_argv() -> Optional[str]:
    parser = argparse.ArgumentParser()
    parser.add_argument("--config-dir", "-cd")
    args, remaining = parser.parse_known_args()
    sys.argv = [sys.argv[0]] + remaining
    return args.config_dir


class ConfigDirSnapshot(Callback):
    _instance = None
    _temp_dir: Optional[Path] = None

    def __new__(cls) -> "ConfigDirSnapshot":
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            cls._temp_dir = None
        return cls._instance

    def create_snapshot_dir(self) -> Optional[str]:
        """Create a snapshot of the config directory by copying all
        configuration files to a temporary directory.

        The snapshot is created by either copying the config directory
        specified by the `--config-dir` argument (and removing it from
        `sys.argv`) or by copying the `conf` directory in the current
        working directory if it exists.
        If no config directory is found, the method returns None.

        Returns:
            The path to the temporary directory containing the snapshot of
            the config directory, or None if no config directory was found.
        """
        if self._temp_dir and self._temp_dir.is_dir():
            return str(self._temp_dir)

        config_dir = pop_config_dir_from_argv()
        if config_dir is None and os.path.isdir("conf"):
            config_dir = "conf"
        if config_dir is None or not os.path.isdir(config_dir):
            return None  # nothing to snapshot

        config_dir = Path(config_dir).resolve()
        self._temp_dir = Path(tempfile.mkdtemp())

        for path in config_dir.rglob("*.yaml"):
            if not path.is_file():
                continue  # skip directories ending with .yaml
            relative_path = path.relative_to(config_dir)
            destination = self._temp_dir / relative_path
            destination.parent.mkdir(parents=True, exist_ok=True)
            shutil.copy2(path, destination)
        return str(self._temp_dir)

    def remove_snapshot_dir(self) -> None:
        """Remove the temporary config directory and all its contents.

        Automatically called at the end of the Hydra multirun.
        """
        try:
            if self._temp_dir and self._temp_dir.is_dir():
                shutil.rmtree(self._temp_dir)
        except Exception as e:
            logging.warning(f"Failed to remove temp config dir: {e}")
        finally:
            self._temp_dir = None  # always try to reset the temp dir

    def on_multirun_end(self, config: DictConfig, **kwargs: Any) -> None:
        """Called in MULTIRUN mode after all jobs returns.

        When using a launcher, this will be executed on local machine.
        """
        self.remove_snapshot_dir()



from .abstract_constants import AbstractConstants
from .export_constants import ExportConstants
from .naming_constants import NamingConstants
from .training_constants import TrainingConstants


__all__ = [
    "AbstractConstants",
    "ExportConstants",
    "NamingConstants",
    "TrainingConstants",
]



from .abstract_script import AbstractScript
from .command_line_error import CommandLineError, print_help_on_error
from .create_script import CreateScript, create
from .delete_failed_script import DeleteFailedScript, rm_failed
from .delete_states_script import DeleteStatesScript, rm_states
from .fetch_script import FetchScript, fetch
from .group_script import GroupScript, group
from .inference_script import InferenceScript, inference
from .list_script import ListScript, list_configs
from .postprocess_script import PostprocessScript, postprocess
from .preprocess_script import PreprocessScript, preprocess
from .show_script import ShowScript, show
from .train_script import TrainScript, train


__all__ = [
    "AbstractScript",
    "CommandLineError",
    "create",
    "CreateScript",
    "DeleteFailedScript",
    "DeleteStatesScript",
    "fetch",
    "FetchScript",
    "group",
    "GroupScript",
    "inference",
    "InferenceScript",
    "list_configs",
    "ListScript",
    "postprocess",
    "PostprocessScript",
    "preprocess",
    "PreprocessScript",
    "print_help_on_error",
    "rm_failed",
    "rm_states",
    "show",
    "ShowScript",
    "train",
    "TrainScript",
]



from typing import Callable, Tuple

import torch

import autrainer
from autrainer.core.structs import AbstractDataBatch
from autrainer.models import AbstractModel
from autrainer.models.utils import create_model_inputs


class SAM(torch.optim.Optimizer):
    def __init__(
        self,
        params: torch.nn.Module,
        base_optimizer: str,
        rho: float = 0.05,
        adaptive: bool = False,
        **kwargs,
    ):
        """Sharpness Aware Minimization (SAM) optimizer.

        This implementation is adapted from the following repository:
        https://github.com/davda54/sam

        For more information, see:
        https://arxiv.org/abs/2010.01412

        Args:
            params: Model parameters.
            base_optimizer: Underlying optimizer performing the sharpness-aware
                update, specified as a relative import path.
            rho: Size of the neighborhood for computing the max loss.
                Defaults to 0.05.
            adaptive: Whether to use an experimental implementation of
                element-wise Adaptive SAM. Defaults to False.
            **kwargs: Additional arguments passed to the underlying optimizer.
        """
        if rho < 0.0:
            raise ValueError(f"Rho '{rho}' should be non-negative.")

        defaults = dict(rho=rho, adaptive=adaptive, **kwargs)
        super(SAM, self).__init__(params, defaults)

        self.base_optimizer = autrainer.instantiate(
            config={
                "_target_": base_optimizer,
                "params": self.param_groups,
                **kwargs,
            },
            instance_of=torch.optim.Optimizer,
        )
        self.param_groups = self.base_optimizer.param_groups
        self.defaults.update(self.base_optimizer.defaults)

    @torch.no_grad()
    def first_step(self, zero_grad=False):
        grad_norm = self._grad_norm()
        for group in self.param_groups:
            scale = group["rho"] / (grad_norm + 1e-12)
            for p in group["params"]:
                if p.grad is None:
                    continue
                self.state[p]["old_p"] = p.data.clone()
                e_w = (
                    (torch.pow(p, 2) if group["adaptive"] else 1.0)
                    * p.grad
                    * scale.to(p)
                )
                p.add_(e_w)  # climb to the local maximum "w + e(w)"

        if zero_grad:
            self.zero_grad()

    # @torch.no_grad()
    # def first_step_uphill(self, zero_grad=False):
    #     grad_norm = self._grad_norm()
    #     for group in self.param_groups:
    #         scale = group["rho"] / (grad_norm + 1e-12)

    #         for p in group["params"]:
    #             if p.grad is None: continue
    #             self.state[p]["old_p"] = p.data.clone()
    #             e_w = (torch.pow(p, 2) if group["adaptive"] else 1.0) * p.grad * scale.to(p)
    #             p.add_(e_w)  # climb to the local maximum "w + e(w)"

    #     if zero_grad: self.zero_grad()

    @torch.no_grad()
    def second_step(self, zero_grad=False):
        for group in self.param_groups:
            for p in group["params"]:
                if p.grad is None:
                    continue
                # get back to "w" from "w + e(w)"
                p.data = self.state[p]["old_p"]

        self.base_optimizer.step()  # do the actual "sharpness-aware" update

        if zero_grad:
            self.zero_grad()

    @torch.no_grad()
    def step(self, closure=None):
        assert (
            closure is not None
        ), "Sharpness Aware Minimization requires closure, but it was not provided"
        # the closure should do a full forward-backward pass
        closure = torch.enable_grad()(closure)

        self.first_step(zero_grad=True)
        closure()
        self.second_step()

    def _grad_norm(self):
        # put everything on the same device, in case of model parallelism
        shared_device = self.param_groups[0]["params"][0].device
        norm = torch.norm(
            torch.stack(
                [
                    ((torch.abs(p) if group["adaptive"] else 1.0) * p.grad)
                    .norm(p=2)
                    .to(shared_device)
                    for group in self.param_groups
                    for p in group["params"]
                    if p.grad is not None
                ]
            ),
            p=2,
        )
        return norm

    def load_state_dict(self, state_dict):
        super().load_state_dict(state_dict)
        self.base_optimizer.param_groups = self.param_groups

    # ? Training Tips functions from the Github repo ref: https://github.com/davda54/sam
    # TODO: Check if Batch Norm Tip needs to be applied
    def custom_step(
        self,
        model: AbstractModel,
        data: AbstractDataBatch,
        criterion: torch.nn.Module,
        probabilities_fn: Callable,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """Sharpness Aware Minimization requires two forward-backward passes
        over the batch to simultaneously minimize the loss value and the loss
        sharpness.

        Args:
            model: The model to train.
            data: The data batch containing features, target, and potentially
                additional fields. The data batch is expected to be on the same
                device as the model. Additional fields are passed to the model
                as keyword arguments if they are present in the model's forward
                method.
            criterion: Loss function.
            probabilities_fn: Function to convert model outputs to
                probabilities.

        Returns:
            Tuple containing the non-reduced loss and model outputs.
        """
        output = model(**create_model_inputs(model, data))
        loss = criterion(probabilities_fn(output), data.target)
        loss.mean().backward()
        self.first_step(zero_grad=True)

        output = model(**create_model_inputs(model, data))
        loss = criterion(probabilities_fn(output), data.target)
        loss.mean().backward()
        self.second_step(zero_grad=True)
        return loss, output



import argparse
import sys
from typing import Callable, NoReturn


class CommandLineError(Exception):
    """Command line error.

    Args:
        parser (argparse.ArgumentParser): Argument parser.
        message (str): Error message.
        code (int, optional): Exit code. Defaults to 2 in accordance with the
        UNIX standard to indicate incorrect usage.
    """

    def __init__(
        self,
        parser: argparse.ArgumentParser,
        message: str = "",
        code: int = 2,
    ) -> None:
        super().__init__(message)
        self.parser = parser
        self.message = message
        self.help = help
        self.code = code

    def handle(self) -> NoReturn:
        """Handle the error and exit."""
        if self.code == 2:
            self.parser.print_help(sys.stderr)
            print(file=sys.stderr)
        if self.message:
            print(self.message, file=sys.stderr)
        sys.exit(self.code)


def print_help_on_error(
    parser: argparse.ArgumentParser,
) -> Callable[[str], NoReturn]:
    """Print help on error.

    Args:
        parser (argparse.ArgumentParser): Argument parser.

    Returns:
        Callable[[str], NoReturn]: Closure that raises a CommandLineError with
        the given message.
    """

    def closure(message: str = "") -> NoReturn:
        if not message:
            raise CommandLineError(parser)

        message = message.capitalize()
        if not message.endswith("."):
            message += "."
        raise CommandLineError(parser, message)

    return closure



import os
from typing import Dict, List, Optional

import numpy as np
import pandas as pd
import torch
import torch.nn.common_types

from autrainer.metrics import AbstractMetric


def disaggregated_evaluation(
    targets: np.ndarray,
    predictions: np.ndarray,
    indices: np.ndarray,
    groundtruth: pd.DataFrame,
    metrics: List[AbstractMetric],
    target_column: List[str],
    stratify: List[str] = None,
) -> Dict:
    r"""Runs evaluation, optionally disaggregated.

    Computes each metric globally (over all targets)
    and unitary (over each target).
    Additionally supports disaggregated evaluations
    for different values
    of columns present in the data dataframe.

    Args:
        targets: array with groundtruth values.
        predictions: array with model predictions.
        indices: array tracking initial dataset indices.
        groundruth: dataframe with groundtruth data and metadata.
        metrics: list of metrics to use for evaluation.
        target_column: columns to evaluate on.
        stratify: optional list of metadata to run evaluation
            in stratified manner.

    Returns:
        Dictionary containing the results of the disaggregated evaluation.

    """
    results = {m.name: {} for m in metrics}
    for metric in metrics:
        if isinstance(target_column, list):
            # this handles the case of multi-label classification and
            # multi-target regression
            # loops over all targets and computes the metric for them
            for idx, col in enumerate(target_column):
                results[metric.name][col] = metric.unitary(
                    targets[:, idx],
                    predictions[:, idx],
                )
        results[metric.name]["all"] = metric(targets, predictions)
        for s in stratify:
            if isinstance(target_column, list):
                raise ValueError(
                    "Stratified evaluation not supported for multi-label "
                    "classification and multi-target regression."
                )
            for v in groundtruth[s].unique():
                idx = groundtruth.loc[groundtruth[s] == v].index
                # Map groundtruth indices to tracker indices
                # This accounts for random shuffling
                mapped_indices = [i for i, x in enumerate(indices) if x in idx]
                results[metric.name][v] = metric(
                    targets[mapped_indices],
                    predictions[mapped_indices],
                )
    return results


def format_results(
    results: Dict[str, float],
    results_type: str,
    training_type: str,
    iteration: Optional[int] = None,
) -> str:
    s = f"{results_type} results"
    s += f" at {training_type} {iteration}" if iteration else ""
    s += ":\n"
    max_key_len = max([len(k) for k in results.keys()])
    s += "\n".join(
        [f"{(k+':').ljust(max_key_len+1)} {v:.4f}" for k, v in results.items()]
    )
    return s


def load_pretrained_model_state(
    model: torch.nn.Module,
    state_dict: Dict[str, torch.Tensor],
    skip_last_layer: bool = True,
) -> None:
    """Load a pretrained model state dict for a model.

    Args:
        model: Model to load the state dict into.
        state_dict: State dict to load into the model.
        skip_last_layer: Whether to skip loading the state dict of the
            last linear or convolutional layer. Defaults to True.
    Raises:
        RuntimeError: If the shapes of the model and state dict do not match.
    """
    last_layer = None
    strict = True
    for name, layer in model.named_modules():
        if isinstance(layer, (torch.nn.Linear, torch.nn.modules.conv._ConvNd)):
            last_layer = name

    if last_layer is not None and skip_last_layer:
        state_dict.pop(last_layer + ".weight", None)
        state_dict.pop(last_layer + ".bias", None)
        strict = False

    model.load_state_dict(state_dict, strict=strict)


def load_pretrained_optim_state(
    optim: torch.optim.Optimizer,
    state_dict: Dict[str, torch.Tensor],
    skip_last_layer: bool = True,
) -> None:
    """Load a pretrained optimizer state dict for an optimizer.

    Args:
        optim: Optimizer to load the state dict into.
        state_dict: State dict to load into the optimizer.
        skip_last_layer: Whether to skip loading the state dict of the
            last layer. Defaults to True.
    """
    if skip_last_layer:
        keys = list(state_dict["state"].keys())
        state_dict["state"].pop(keys[-1], None)  # bias
        state_dict["state"].pop(keys[-2], None)  # weight

    optim.load_state_dict(state_dict)


def load_checkpoint(checkpoint: str) -> Dict[str, torch.Tensor]:
    """Load a checkpoint state dict from a file.

    Args:
        checkpoint: Path to the checkpoint file.

    Raises:
        FileNotFoundError: If the checkpoint file does not exist.

    Returns:
        The model state dict.
    """
    # TODO: Support Hugging Face Checkpoints in the future.
    if not os.path.isfile(checkpoint):
        raise FileNotFoundError(f"Checkpoint file not found: {checkpoint}")
    return torch.load(checkpoint)



from functools import cached_property
import os
from typing import Dict, List, Optional, Union

from omegaconf import DictConfig
import pandas as pd

from autrainer.transforms import SmartCompose

from .abstract_dataset import BaseClassificationDataset


class AIBO(BaseClassificationDataset):
    def __init__(
        self,
        path: str,
        features_subdir: Optional[str],
        seed: int,
        metrics: List[Union[str, DictConfig, Dict]],
        tracking_metric: Union[str, DictConfig, Dict],
        index_column: str,
        target_column: str,
        file_type: str,
        file_handler: Union[str, DictConfig, Dict],
        features_path: Optional[str] = None,
        train_transform: Optional[SmartCompose] = None,
        dev_transform: Optional[SmartCompose] = None,
        test_transform: Optional[SmartCompose] = None,
        stratify: Optional[List[str]] = None,
        aibo_task: str = "2cl",
    ) -> None:
        """FAU AIBO dataset.

        Args:
            path: Root path to the dataset.
            features_subdir: Subdirectory containing the features.
                If `None`, defaults to audio subdirectory,
                which is `default` for the standard format,
                but can be overridden in the dataset specification.
            seed: Seed for reproducibility.
            metrics: List of metrics to calculate.
            tracking_metric: Metric to track.
            index_column: Index column of the dataframe.
            target_column: Target column of the dataframe.
            file_type: File type of the features.
            file_handler: File handler to load the data.
            features_path: Root path to features. Useful
                when features need to be extracted and stored
                in a different directory than the root of the dataset.
                If `None`, will be set to `path`. Defaults to `None`.
            train_transform: Transform to apply to the training set.
                Defaults to None.
            dev_transform: Transform to apply to the development set.
                Defaults to None.
            test_transform: Transform to apply to the test set.
                Defaults to None.
            stratify: Columns to stratify the dataset on. Defaults to None.
            aibo_task: Task to load in ["2cl", "5cl"]. Defaults to "2cl".
        """
        self._assert_choice(aibo_task, ["2cl", "5cl"], "aibo_task")
        self.aibo_task = aibo_task
        super().__init__(
            path=path,
            features_subdir=features_subdir,
            seed=seed,
            metrics=metrics,
            tracking_metric=tracking_metric,
            index_column=index_column,
            target_column=target_column,
            file_type=file_type,
            file_handler=file_handler,
            features_path=features_path,
            train_transform=train_transform,
            dev_transform=dev_transform,
            test_transform=test_transform,
            stratify=stratify,
        )

    @property
    def audio_subdir(self) -> str:
        """Subdirectory containing audio data."""
        return "wav"

    @cached_property
    def _load_df(self) -> pd.DataFrame:
        df = pd.read_csv(
            os.path.join(
                self.path, f"chunk_labels_{self.aibo_task}_corpus.txt"
            ),
            header=None,
            sep=" ",
        )
        df = df.rename(columns={0: "id", 1: "class", 2: "conf"})
        df["file"] = df["id"].apply(lambda x: x + ".wav")
        df["school"] = df["id"].apply(lambda x: x.split("_")[0])
        df["speaker"] = df["id"].apply(lambda x: x.split("_")[1])
        return df

    @cached_property
    def df_train(self) -> pd.DataFrame:
        df = self._load_df
        df_train_dev = df.loc[df["school"] == "Ohm"]
        speakers = sorted(df_train_dev["speaker"].unique())
        df_train = df_train_dev.loc[
            df_train_dev["speaker"].isin(speakers[:-2])
        ]
        return df_train

    @cached_property
    def df_dev(self) -> pd.DataFrame:
        df = self._load_df
        df_train_dev = df.loc[df["school"] == "Ohm"]
        speakers = sorted(df_train_dev["speaker"].unique())
        df_dev = df_train_dev.loc[df_train_dev["speaker"].isin(speakers[-2:])]
        return df_dev

    @cached_property
    def df_test(self) -> pd.DataFrame:
        df = self._load_df
        df_test = df.loc[df["school"] == "Mont"]
        return df_test

    @staticmethod
    def download(path: str) -> None:  # pragma: no cover
        """
        Download the FAU AIBO dataset.

        As the AIBO dataset is private, this method does not download the
        dataset but rather prepares the file structure expected by the
        preprocessing routines.

        In the specified path, the following directories and files are expected:

        - `default/`: Directory containing .wav files.
        - `chunk_labels_2cl_corpus.txt`: File containing the file names and
          corresponding labels for the 2-class classification task.
        - `chunk_labels_5cl_corpus.txt`: File containing the file names and
          corresponding labels for the 5-class classification task.

        For more information on the dataset and dataset split, see:
        https://doi.org/10.1109/ICME51207.2021.9428217

        Args:
            path: Path to the directory to download the dataset to.
        """
        if not os.path.isfile(
            os.path.join(path, "chunk_labels_2cl_corpus.txt")
            or os.path.isfile(
                os.path.join(path, "chunk_labels_5cl_corpus.txt")
            )
        ):
            raise ValueError(
                f"File 'chunk_labels_2cl_corpus.txt' or "
                f"'chunk_labels_5cl_corpus.txt' does not exist in '{path}'."
            )



import random

import numpy as np
import torch


def set_seed(seed: int) -> None:
    """Set a global seed for reproducibility for random, numpy, and torch.

    If CUDA is available, set the seed for CUDA and cuDNN as well.

    Args:
        seed: Seed to set.
    """
    if seed != seed % 2**32:
        raise ValueError(f"Seed must be between 0 and 2**32-1, got '{seed}'.")
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    if torch.cuda.is_available():
        torch.cuda.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = False



from abc import ABC
from typing import Any, Type


class AbstractConstants(ABC):
    """Abstract constants singleton class for managing the configurations of
    `autrainer`.
    """

    _name = "AbstractConstants"
    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance

    def _assert_type(
        self,
        value: Any,
        value_type: Type[Any],
        constant: str,
        msg: str = "",
    ) -> None:
        if not isinstance(value, value_type):
            msg = f" {msg}" if msg else ""
            raise ValueError(
                f"Invalid type for '{self._name}.{constant}'{msg}: "
                f"expected '{value_type.__name__}', "
                f"but got '{type(value).__name__}' with value '{value}'."
            )



r"""Code taken from https://github.com/end2you/end2you.

Copied only model code as needed.
Adapted to match coding styles of current repo:
- Remove input_size argument
+ Add output_dim argument
+ Changed RNN model to local Sequential
+ Added time_pooling layer before linear
  + This changes model architecture
+ Remove Emo16
  + Seems broken; 2nd max-pool applied over channels
  + Output dimension is too long
  + Model's results never reproduced anyway

"""

from typing import Optional, Tuple, Union

import numpy as np
import torch
import torch.nn as nn

from .abstract_model import AbstractModel
from .sequential import Sequential
from .utils import assert_no_transfer_weights


class Base(nn.Module):
    """Base class to build convolutional neural network model."""

    def __init__(
        self,
        conv_layers_args: dict,
        maxpool_layers_args: dict,
        conv_op: nn = nn.Conv1d,
        max_pool_op: nn = nn.MaxPool1d,
        activ_fn: nn = nn.LeakyReLU(),
        normalize: bool = False,
    ) -> None:
        """Audio model.

        Args:
            conv_layers_args: Parameters of convolutions layers.
            maxpool_layers_args: Parameters of max pool layer layers.
            conv_op: Convolution operation to use. Defaults to torch.nn.Conv1d.
            max_pool_op: Max pooling operation to use.
                Defaults to torch.nn.MaxPool1d.
            activ_fn: Activation function to use.
                Defaults to torch.nn.LeakyReLU().
            normalize: Whether to use batch normalization. Defaults to False.
        """

        super().__init__()
        self.conv_layers_args = conv_layers_args
        self.maxpool_layers_args = maxpool_layers_args
        self.conv_op = conv_op
        self.max_pool_op = max_pool_op
        self.activ_fn = activ_fn
        self.normalize = normalize

        network_layers = nn.ModuleList()
        for conv_args, mp_args in zip(
            *[conv_layers_args.values(), maxpool_layers_args.values()]
        ):
            network_layers.extend(
                [self._conv_block(conv_args, activ_fn, normalize)]
            )
            network_layers.extend([max_pool_op(**mp_args)])

        self.network = nn.Sequential(*network_layers)
        self.reset_parameters()

    def reset_parameters(self) -> None:
        """Initialize parameters of the model."""
        for m in list(self.modules()):
            self._init_weights(m)

    def _init_weights(self, m: nn.Module) -> None:
        """Helper method to initialize the parameters of the model
        with Kaiming uniform initialization.

        Args:
            m: Module to initialize.
        """

        if isinstance(m, nn.Conv1d) or isinstance(m, nn.Linear):
            nn.init.kaiming_uniform_(m.weight)
            nn.init.zeros_(m.bias)
        if isinstance(m, nn.LSTM):
            for name, param in m.named_parameters():
                if "bias" in name:
                    nn.init.zeros_(param)
                elif "weight" in name:
                    nn.init.kaiming_uniform_(param)

    @classmethod
    def _num_out_features(
        cls,
        input_size: int,
        conv_args: dict,
        mp_args: dict,
    ) -> int:
        """Number of features extracted from Convolution Neural Network.

        Args:
            input_size: Number of samples of the frame.
            conv_args: Parameters of convolutions layers.
            mp_args: Parameters of max pool layer layers.

        Returns:
            Number of features extracted from the network layers.
        """

        layer_input = input_size
        for i, (conv_arg, mp_arg) in enumerate(
            zip(*[conv_args.values(), mp_args.values()])
        ):
            # number of features in the convolution output
            layer_input = np.floor(
                (
                    layer_input
                    - conv_arg["kernel_size"]
                    + 2 * conv_arg["padding"]
                )
                / conv_arg["stride"]
                + 1
            )

            layer_input = np.floor(
                (layer_input - mp_arg["kernel_size"]) / mp_arg["stride"] + 1
            )

        return int(layer_input)

    def _conv_block(
        self,
        conv_args: dict,
        activ_fn: nn,
        normalize: bool = False,
    ) -> nn.Module:
        """Convolution block.

        Args:
            conv_args: Parameters of convolution layer.
            activ_fn: Activation function to use. Defaults to
                torch.nn.LeakyReLU().
            normalize: Whether to use batch normalization. Defaults to False.
        """

        layer = nn.ModuleList([self.conv_op(**conv_args)])

        if normalize:
            layer.append(nn.BatchNorm1d(conv_args["out_channels"]))

        layer.append(activ_fn)
        return nn.Sequential(*layer)

    def embeddings(self, x: torch.Tensor) -> torch.Tensor:
        return self.network(x)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.embeddings(x)


class Emo18(nn.Module):
    def __init__(self) -> None:
        """Speech emotion recognition model proposed in:
        https://doi.org/10.1109/ICASSP.2018.8462677
        """

        super().__init__()
        self.model, self.num_features = self.build_audio_model()

    def build_audio_model(self) -> Tuple[nn.Module, int]:
        """Build the audio model: 3 blocks of convolution + max-pooling."""

        out_channels = [64, 128, 256]
        in_channels = [1]
        in_channels.extend([x for x in out_channels[:-1]])
        kernel_size = [8, 6, 6]
        stride = [1, 1, 1]
        padding = ((np.array(kernel_size) - 1) // 2).tolist()

        num_layers = len(in_channels)
        conv_args = {
            f"layer{i}": {
                "in_channels": in_channels[i],
                "out_channels": out_channels[i],
                "kernel_size": kernel_size[i],
                "stride": stride[i],
                "padding": padding[i],
            }
            for i in range(num_layers)
        }

        kernel_size = [10, 8, 8]
        stride = [10, 8, 8]
        maxpool_args = {
            f"layer{i}": {"kernel_size": kernel_size[i], "stride": stride[i]}
            for i in range(num_layers)
        }

        audio_model = Base(conv_args, maxpool_args, normalize=True)
        # conv_red_size = Base._num_out_features(
        #     input_size, conv_args, maxpool_args)
        num_layers = len(in_channels) - 1
        # num_out_features = conv_red_size * \
        #     conv_args[f"layer{num_layers}"]["out_channels"]

        num_out_features = conv_args[f"layer{num_layers}"]["out_channels"]

        return audio_model, num_out_features

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.model(x)


class Zhao19(nn.Module):
    def __init__(self) -> None:
        """Speech emotion recognition model proposed in:
        https://doi.org/10.1016/j.bspc.2018.08.035
        """

        super().__init__()
        self.model, self.num_features = self.build_audio_model()

    def build_audio_model(self) -> Tuple[nn.Module, int]:
        """Build the audio model: 3 blocks of convolution + max-pooling."""

        out_channels = [64, 64, 128, 128]
        in_channels = [1]
        in_channels.extend([x for x in out_channels[:-1]])
        kernel_size = [3, 3, 3, 3]
        stride = [1, 1, 1, 1]
        padding = ((np.array(kernel_size) - 1) // 2).tolist()

        num_layers = len(in_channels)
        conv_args = {
            f"layer{i}": {
                "in_channels": in_channels[i],
                "out_channels": out_channels[i],
                "kernel_size": kernel_size[i],
                "stride": stride[i],
                "padding": padding[i],
            }
            for i in range(num_layers)
        }

        kernel_size = [4, 4, 4, 4]
        stride = [4, 4, 4, 4]
        maxpool_args = {
            f"layer{i}": {"kernel_size": kernel_size[i], "stride": stride[i]}
            for i in range(num_layers)
        }

        audio_model = Base(
            conv_args, maxpool_args, normalize=True, activ_fn=nn.ELU()
        )

        return audio_model, out_channels[-1]

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.model(x)


class AudioModel(nn.Module):
    def __init__(
        self,
        model_name: str,
        *args,
        **kwargs,
    ):
        """Audio network model.

        Args:
            model_name: Name of the model in ["emo18", "zhao19"].
            *args: Additional arguments to the model.
            **kwargs: Additional keyword arguments to the model.
        """

        super().__init__()

        self.model = self._get_model(model_name)
        self.model = self.model(*args, **kwargs)
        self.num_features = self.model.num_features

    def _get_model(self, model_name: str) -> nn.Module:
        """Factory method to choose audio model."""

        return {"emo18": Emo18, "zhao19": Zhao19}[model_name]

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.model(x)


class AudioRNNModel(AbstractModel):
    def __init__(
        self,
        output_dim: int,
        model_name: str,
        hidden_size: int = 256,
        num_layers: int = 2,
        dropout: float = 0.5,
        cell: str = "LSTM",
        bidirectional: bool = False,
        transfer: Optional[Union[bool, str]] = None,
    ) -> None:
        """Audio RNN model.

        Args:
            output_dim: Output dimension of the model.
            model_name: Model name in ["emo18", "zhao19"].
            hidden_size: Hidden size of the RNN. Defaults to 256.
            num_layers: Number of layers of the RNN. Defaults to 2.
            dropout: Dropout rate. Defaults to 0.5.
            cell: Type of RNN cell in ["LSTM", "GRU"] Defaults to "LSTM".
            bidirectional: Whether to use a bidirectional RNN.
                Defaults to False.
            transfer: Not available for this model. If set, raises an error.
                Defaults to None.
        """
        assert_no_transfer_weights(self, transfer)
        super().__init__(output_dim, None)  # no transfer learning weights
        self.model_name = model_name
        self.hidden_size = hidden_size
        self.num_layers = num_layers
        self.dropout = dropout
        self.cell = cell
        self.bidirectional = bidirectional
        audio_network = AudioModel(model_name=model_name)
        self.audio_model = audio_network
        self.rnn = Sequential(
            input_dim=self.audio_model.num_features,
            hidden_size=hidden_size,
            num_layers=num_layers,
            dropout=dropout,
            cell=cell,
            time_pooling=True,
            bidirectional=bidirectional,
        )
        self.linear = nn.Linear(self.rnn.hidden_size, self.output_dim)

    def embeddings(self, features: torch.Tensor) -> torch.Tensor:
        batch_size, seq_length, t = features.shape
        features = features.view(batch_size * seq_length, 1, t)
        audio_out = self.audio_model(features)
        audio_out = audio_out.transpose(1, 2)

        rnn_out = self.rnn(audio_out)

        return rnn_out

    def forward(self, features: torch.Tensor) -> torch.Tensor:
        """
        Args:
            features ((torch.Tensor) - BS x S x 1 x T)
        """
        rnn_out = self.embeddings(features)
        output = self.linear(rnn_out)
        return output



from dataclasses import dataclass
from typing import List, Optional

from autrainer.core.constants import NamingConstants
from autrainer.core.scripts.abstract_script import MockParser

from .abstract_postprocess_script import (
    AbstractPostprocessArgs,
    AbstractPostprocessScript,
)
from .command_line_error import CommandLineError
from .utils import catch_cli_errors


@dataclass
class PostprocessArgs(AbstractPostprocessArgs):
    max_runs: Optional[int]
    aggregate: Optional[List[List[str]]]


class PostprocessScript(AbstractPostprocessScript):
    def __init__(self) -> None:
        super().__init__(
            "postprocess",
            "Postprocess grid search results.",
            epilog="Example: autrainer postprocess results default -a seed",
            dataclass=PostprocessArgs,
        )

    def add_arguments(self) -> None:
        super().add_arguments()
        self.parser.add_argument(
            "-m",
            "--max-runs",
            type=int,
            metavar="N",
            required=False,
            help="Maximum number of best runs to plot.",
            default=None,
        )
        self.parser.add_argument(
            "-a",
            "--aggregate",
            type=str,
            metavar="A",
            required=False,
            nargs="+",
            action="append",
            help="Configurations to aggregate. One or more of:"
            + "\n - ".join(
                [""] + sorted(NamingConstants().VALID_AGGREGATIONS)
            ),
        )

    def main(self, args: PostprocessArgs) -> None:
        super().main(args)
        self._assert_valid_aggregations(args)
        self._summarize(args)

    def _assert_valid_aggregations(self, args: PostprocessArgs) -> None:
        if not args.aggregate:
            return
        for agg in args.aggregate:
            for a in agg:
                if a not in NamingConstants().VALID_AGGREGATIONS:
                    m = f"Invalid configuration to aggregate: '{a}'."
                    raise CommandLineError(self.parser, m)

    def _summarize(self, args: PostprocessArgs) -> None:
        from autrainer.postprocessing import (
            AggregateGrid,
            SummarizeGrid,
        )

        sg = SummarizeGrid(
            results_dir=args.results_dir,
            experiment_id=args.experiment_id,
            max_runs_plot=args.max_runs,
        )
        sg.summarize()
        sg.plot_aggregated_bars()
        sg.plot_metrics()

        if not args.aggregate:
            return
        for agg in args.aggregate:
            ag = AggregateGrid(
                results_dir=args.results_dir,
                experiment_id=args.experiment_id,
                aggregate_list=agg,
                max_runs_plot=args.max_runs,
            )
            ag.aggregate()
            ag.summarize()


@catch_cli_errors
def postprocess(
    results_dir: str,
    experiment_id: str,
    max_runs: Optional[int] = None,
    aggregate: Optional[List[List[str]]] = None,
) -> None:
    """Postprocess grid search results.

    If called in a notebook, the function will not raise an error and print
    the error message instead.

    Args:
        results_dir: Path to grid search results directory.
        experiment_id: ID of experiment to postprocess.
        max_runs: Maximum number of best runs to plot. Defaults to None.
        aggregate: Configurations to aggregate. One or more of:
            :const:`~autrainer.core.constants.NamingConstants.VALID_AGGREGATIONS`.
            Defaults to None.

    Raises:
        CommandLineError: If the results directory or experiment ID dont exist.
    """
    script = PostprocessScript()
    script.parser = MockParser()
    script.main(
        PostprocessArgs(
            results_dir,
            experiment_id,
            max_runs,
            aggregate,
        )
    )



import numpy as np


def assert_nonzero_frequency(frequency: np.ndarray, num_targets: int) -> None:
    """Check if the frequency array contains non-zero values for all classes
    and has the correct length and therefore does not contain any missing
    (zero) class-frequencies.

    Args:
        frequency: Array with the frequency of each class.
        num_targets: Number of target classes.

    Raises:
        ValueError: If the frequency array contains zero values or has an
            incorrect length.
    """
    if (frequency == 0).any() or len(frequency) != num_targets:
        raise ValueError(
            "Balanced weighting requires a non-zero frequency for all classes."
        )



import numpy as np
import sklearn.metrics

from .abstract_metric import BaseAscendingMetric


class MLAccuracy(BaseAscendingMetric):
    def __init__(self):
        """Accuracy metric using `sklearn.metrics.accuracy_score`."""
        super().__init__(name="ml-accuracy", fn=sklearn.metrics.accuracy_score)


class MLF1Macro(BaseAscendingMetric):
    def __init__(self):
        """F1 macro metric using `sklearn.metrics.f1_score`."""
        super().__init__(
            name="ml-f1-macro",
            fn=sklearn.metrics.f1_score,
            average="macro",
        )

    def unitary(self, y_true: np.ndarray, y_pred: np.ndarray) -> float:
        """Unitary evaluation of metric.

        Metric computed with `average='binary'`
        i.e. only accounting for the positive label.

        Args:
            y_true: ground truth values.
            y_pred: prediction values.

        Returns:
            The unitary score.
        """
        return float(
            sklearn.metrics.f1_score(
                y_true=y_true, y_pred=y_pred, average="binary"
            )
        )


class MLF1Micro(BaseAscendingMetric):
    def __init__(self):
        """F1 micro metric using `sklearn.metrics.f1_score`."""
        super().__init__(
            name="ml-f1-micro",
            fn=sklearn.metrics.f1_score,
            average="micro",
        )

    def unitary(self, y_true: np.ndarray, y_pred: np.ndarray) -> float:
        """Unitary evaluation of metric.

        Metric computed with `average='binary'`
        i.e. only accounting for the positive label.

        Args:
            y_true: ground truth values.
            y_pred: prediction values.

        Returns:
            The unitary score.
        """
        return float(
            sklearn.metrics.f1_score(
                y_true=y_true, y_pred=y_pred, average="binary"
            )
        )


class MLF1Weighted(BaseAscendingMetric):
    def __init__(self):
        """F1 weighted metric using `sklearn.metrics.f1_score`."""
        super().__init__(
            name="ml-f1-weighted",
            fn=sklearn.metrics.f1_score,
            average="weighted",
        )

    def unitary(self, y_true: np.ndarray, y_pred: np.ndarray) -> float:
        """Unitary evaluation of metric.

        Metric computed with `average='binary'`
        i.e. only accounting for the positive label.

        Args:
            y_true: ground truth values.
            y_pred: prediction values.

        Returns:
            The unitary score.
        """
        return float(
            sklearn.metrics.f1_score(
                y_true=y_true, y_pred=y_pred, average="binary"
            )
        )



from typing import Dict, Optional, Tuple, Union

from omegaconf import DictConfig

import autrainer
from autrainer.transforms import SmartCompose

from .augmentation_pipeline import AugmentationPipeline


class AugmentationManager:
    def __init__(
        self,
        train_augmentation: Optional[Union[DictConfig, Dict]] = None,
        dev_augmentation: Optional[Union[DictConfig, Dict]] = None,
        test_augmentation: Optional[Union[DictConfig, Dict]] = None,
    ) -> None:
        """Manage the creation of the augmentation pipelines for train, dev,
        and test sets.

        Args:
            train_augmentation: Train augmentation configuration.
            dev_augmentation: Dev augmentation configuration.
            test_augmentation: Test augmentation configuration.
        """
        self.train = train_augmentation
        self.dev = dev_augmentation
        self.test = test_augmentation

    def get_augmentations(
        self,
    ) -> Tuple[SmartCompose, SmartCompose, SmartCompose]:
        """Get augmentation pipelines for train, dev, and test.

        Returns:
            Tuple of augmentation pipelines for train, dev, and test.
        """
        return (
            self._build_augmentation(self.train),
            self._build_augmentation(self.dev),
            self._build_augmentation(self.test),
        )

    def _build_augmentation(
        self,
        augmentation: Optional[Union[DictConfig, Dict]],
    ) -> SmartCompose:
        if augmentation is None:
            return SmartCompose([])
        pipeline = autrainer.instantiate(
            augmentation,
            instance_of=AugmentationPipeline,
        )
        return pipeline.create_pipeline()



from typing import List

from .abstract_constants import AbstractConstants


class TrainingConstants(AbstractConstants):
    """Singleton for managing the training configurations of `autrainer`."""

    _name = "TrainingConstants"
    _tasks = [
        "classification",
        "ml-classification",
        "regression",
        "mt-regression",
    ]

    @property
    def TASKS(self) -> List[str]:
        """Get the supported training tasks.
        Defaults to :attr:`["classification", "ml-classification",
        "regression", "mt-regression"]`.

        Returns:
            Supported training tasks.
        """
        return self._tasks

    @TASKS.setter
    def TASKS(self, tasks: List[str]):
        """Set the supported training tasks.

        Args:
            training_tasks: Supported training tasks.

        Raises:
            ValueError: If the training tasks are not a list of strings.
        """

        self._assert_type(tasks, list, "TASKS")
        for t in tasks:
            self._assert_type(t, str, "TASKS", "in training tasks")
        self._tasks = tasks



import logging
from typing import Dict, Optional, Union

from omegaconf import DictConfig

from .abstract_logger import AbstractLogger


class FallbackLogger(AbstractLogger):
    def __init__(
        self,
        requested_logger: Optional[str] = None,
        extras: Optional[str] = None,
    ) -> None:
        """Fallback logger for when a requested logger is not available.

        If the requested logger is not available, a warning is logged.
        If both requested_logger and extras are None, nothing is logged.
        The logger serves as a no-op.

        Args:
            requested_logger: The requested logger. Defaults to None.
            extras: The extras required to install the logger. Defaults to None.
        """
        if requested_logger is None and extras is None:
            return
        log = logging.getLogger(__name__)
        log.warning(
            f"Requested logger '{requested_logger}' not available. "
            "Install the required extras with "
            f"'pip install autrainer[{extras}]'."
        )

    def log_params(self, params: Union[dict, DictConfig]) -> None:
        pass

    def log_and_update_metrics(
        self, metrics: Dict[str, Union[int, float]], iteration: int = None
    ) -> None:
        pass

    def log_metrics(
        self,
        metrics: Dict[str, Union[int, float]],
        iteration=None,
    ) -> None:
        pass

    def log_timers(self, timers: Dict[str, float]) -> None:
        pass

    def log_artifact(self, filename: str, path: str = "") -> None:
        pass



"""Code for AudioSet dataset.
Utility functions to filter data
adapted from: https://github.com/audeering/audtorch/blob/master/audtorch/datasets/audio_set.py
Originally written by https://github.com/hagenw
"""

from functools import cached_property
import json
import logging
import os
from typing import Dict, List, Optional, Union

from audeer import flatten_list
from omegaconf import DictConfig
import pandas as pd

from autrainer.datasets.abstract_dataset import BaseMLClassificationDataset
from autrainer.transforms import SmartCompose


class AudioSet(BaseMLClassificationDataset):
    categories = {
        "Human sounds": [
            "Human voice",
            "Whistling",
            "Respiratory sounds",
            "Human locomotion",
            "Digestive",
            "Hands",
            "Heart sounds, heartbeat",
            "Otoacoustic emission",
            "Human group actions",
        ],
        "Source-ambiguous sounds": [
            "Generic impact sounds",
            "Surface contact",
            "Deformable shell",
            "Onomatopoeia",
            "Silence",
            "Other sourceless",
        ],
        "Animal": [
            "Domestic animals, pets",
            "Livestock, farm animals, working animals",
            "Wild animals",
        ],
        "Sounds of things": [
            "Vehicle",
            "Engine",
            "Domestic sounds, home sounds",
            "Bell",
            "Alarm",
            "Mechanisms",
            "Tools",
            "Explosion",
            "Wood",
            "Glass",
            "Liquid",
            "Miscellaneous sources",
            "Specific impact sounds",
        ],
        "Music": [
            "Musical instrument",
            "Music genre",
            "Musical concepts",
            "Music role",
            "Music mood",
        ],
        "Natural sounds": ["Wind", "Thunderstorm", "Water", "Fire"],
        "Channel, environment and background": [
            "Acoustic environment",
            "Noise",
            "Sound reproduction",
        ],
    }

    def __init__(
        self,
        path: str,
        features_subdir: Optional[str],
        seed: int,
        metrics: List[Union[str, DictConfig, Dict]],
        tracking_metric: Union[str, DictConfig, Dict],
        index_column: str,
        file_type: str,
        file_handler: Union[str, DictConfig, Dict],
        target_column: Optional[List[str]] = None,
        features_path: Optional[str] = None,
        train_transform: Optional[SmartCompose] = None,
        dev_transform: Optional[SmartCompose] = None,
        test_transform: Optional[SmartCompose] = None,
        stratify: Optional[List[str]] = None,
        threshold: float = 0.5,
        use_unbalanced: bool = False,
        include: Optional[List[str]] = None,
        exclude: Optional[List[str]] = None,
    ) -> None:
        """AudioSet dataset.

        .. warning::

            AudioSet changes constantly
            as videos are removed from
            YouTube over time.
            Results are not reproducible
            across different snapshots
            of the data.

        Args:
            path: Root path to the dataset.
            features_subdir: Subdirectory containing the features.
                If `None`, defaults to audio subdirectory,
                which is `default` for the standard format,
                but can be overridden in the dataset specification.
            seed: Seed for reproducibility.
            metrics: List of metrics to calculate.
            tracking_metric: Metric to track.
            index_column: Index column of the dataframe.
            target_column: Target column of the dataframe.
                If None, defaults to the least common subset
                available in the data. Defaults to None.
            file_type: File type of the features.
            file_handler: File handler to load the data.
            batch_size: Batch size.
            inference_batch_size: Inference batch size. If None, defaults to
                batch_size. Defaults to None.
            features_path: Root path to features. Useful
                when features need to be extracted and stored
                in a different directory than the root of the dataset.
                If `None`, will be set to `path`. Defaults to `None`.
            train_transform: Transform to apply to the training set.
                Defaults to None.
            dev_transform: Transform to apply to the development set.
                Defaults to None.
            test_transform: Transform to apply to the test set.
                Defaults to None.
            stratify: Columns to stratify the dataset on. Defaults to None.
            threshold: Threshold for classification. Defaults to 0.5.
            use_unbalanced: Flag to allow the use of unbalanced set
                during training.
            include: List of categories to include. If set,
                only instances tagged with at least one
                category from the list will be included
                in the final dataset.
            exclude: List of categories to include. If set,
                all instances tagged with at least one
                category from the list will be excluded
                from the final dataset.
        """
        self.use_unbalanced = use_unbalanced
        self.include = include
        self.exclude = exclude
        self._log = logging.getLogger(__name__)
        super().__init__(
            path=path,
            features_subdir=features_subdir,
            seed=seed,
            metrics=metrics,
            tracking_metric=tracking_metric,
            index_column=index_column,
            target_column=target_column,
            file_type=file_type,
            file_handler=file_handler,
            features_path=features_path,
            train_transform=train_transform,
            dev_transform=dev_transform,
            test_transform=test_transform,
            stratify=stratify,
            threshold=threshold,
        )

    @property
    def audio_subdir(self) -> str:
        """Subdirectory containing audio data.

        Data assumed to be in root directory
        under the name of the respective partition
        which is prepended once the CSV file is loaded.
        Available partitions are (with original number of instances):
        - balanced_train_segments (n=22,160)
        - unbalanced_train_segments (n=2,041,789)
        - eval_segments (n=20,371)
        """
        return ""

    def _assert_target_column(self, allowed_columns: List[str]) -> None:
        """Override target column check.

        Allow `None` as a valid input.
        Override it to support all available tags
        in ontology, irrespective of whether
        they appear at all in the data or not.
        """
        if self.target_column is None:
            unique_categories = set([x["name"] for x in self.ontology])
            train_categories = set(
                [
                    x
                    for y in self.df_train["categories"].values.tolist()
                    for x in y
                ]
            )
            dev_categories = set(
                [
                    x
                    for y in self.df_dev["categories"].values.tolist()
                    for x in y
                ]
            )
            self.target_column = list(
                (
                    train_categories.intersection(dev_categories).intersection(
                        unique_categories
                    )
                )
            )
            self._log.info(
                (
                    f"Training on {len(self.target_column)} classes:\n"
                    f"{self.target_column}"
                )
            )
        return super()._assert_target_column(allowed_columns)

    def _filename_and_ids(self, df: pd.DataFrame) -> pd.DataFrame:
        r"""Return data frame with filenames and IDs.

        Args:
            df: data frame as read in from the CSV file.

        Results:
            data frame with columns `filename` and `ids`.

        """
        df = df.rename(columns={"positive_labels": "ids"})
        # Translate labels from "label1,label2" to [label1, label2]
        df["ids"] = [label.strip('"').split(",") for label in df["ids"]]
        # Insert filename
        df["filename"] = df["# YTID"] + ".wav"
        return df[["filename", "ids"]]

    def _add_parent_ids(self, child_ids: List[str]) -> List[str]:
        r"""Add all parent IDs to the list of given child IDs.

        Args:
            child_ids: child IDs.

        Return:
            list of child and parent IDs.

        """
        ids = child_ids
        for id in child_ids:
            ids += [x["id"] for x in self.ontology if id in x["child_ids"]]
        # Remove duplicates
        return list(set(ids))

    def _convert_ids_to_categories(self, ids: List[str]) -> List[str]:
        r"""Convert list of ids to sorted list of categories.

        Args:
            ids: list of IDs.

        Returns:
            list of sorted categories.

        """
        # Convert IDs to categories
        categories = []
        for id in ids:
            categories += [x["name"] for x in self.ontology if x["id"] == id]
        # Order categories after the first two top ontologies
        order = []
        first_hierarchy = self.categories.keys()
        second_hierarchy = flatten_list(list(self.categories.values()))
        for cat in categories:
            if cat in first_hierarchy:
                order += [0]
            elif cat in second_hierarchy:
                order += [1]
            else:
                order += [2]
        # Sort list `categories` by the list `order`
        categories = [cat for _, cat in sorted(zip(order, categories))]
        return categories

    def _filter_by_categories(
        self,
        df: pd.DataFrame,
        categories: List[str],
        exclude_mode: bool = False,
    ) -> pd.DataFrame:
        r"""Return data frame containing only specified categories.

        Args:
            df: data frame containing the columns `ids`.
            categories: list of categories to include or exclude.
            exclude_mode: if `False` the specified categories
                should be included in the data frame, otherwise excluded.
                Default: `False`.

        Returns:
            data frame containing only the desired categories.

        """
        ids = self._ids_for_categories(categories)
        if exclude_mode:
            # Remove rows that have an intersection of actual and desired IDs
            df = df.loc[
                df.apply(
                    lambda row: False if set(row["ids"]) & set(ids) else True,
                    axis=1,
                )
            ]
        else:
            # Include rows that have an intersection of actual and desired IDs
            df = df.loc[
                df.apply(
                    lambda row: True if set(row["ids"]) & set(ids) else False,
                    axis=1,
                )
            ]
        df = df.reset_index(drop=True)
        return df

    def _ids_for_categories(self, categories: List[str]) -> List[str]:
        r"""All IDs and child IDs for a given set of categories.

        Args:
            categories: list of categories.

        Returns:
            list of IDs.

        """
        ids = []
        category_ids = [
            x["id"] for x in self.ontology if x["name"] in categories
        ]
        for category_id in category_ids:
            ids += self._subcategory_ids(category_id)
        # Remove duplicates
        return list(set(ids))

    def _subcategory_ids(self, parent_id: str) -> List[str]:
        r"""Recursively identify all IDs of a given category.

        Args:
            parent_id: ID of parent category

        Returns:
            list of all children IDs and the parent ID

        """
        id_list = [parent_id]
        child_ids = [
            x["child_ids"] for x in self.ontology if x["id"] == parent_id
        ]
        child_ids = flatten_list(child_ids)
        # Add all subcategories
        for child_id in child_ids:
            id_list += self._subcategory_ids(child_id)
        return id_list

    def map_to_classes(self, df: pd.DataFrame) -> pd.DataFrame:
        """Map dataframe as loaded to `autrainer` compatible.

        Assigns all AudioSet IDs to category names.
        Recursively assigns parent nodes
        to every instance where a child is present.
        Optionally filters dataframes
        to include/exclude instances.
        Creates columns with final labels.

        Args:
            df: dataframe loaded from CSV.

        Returns:
            Adapted dataframe.
        """
        df = self._filename_and_ids(df)
        if self.include is not None:
            df = self._filter_by_categories(df, self.include)
        df["ids"] = df["ids"].map(self._add_parent_ids)
        if self.exclude is not None:
            df = self._filter_by_categories(
                df, self.exclude, exclude_mode=True
            )
        df["categories"] = df["ids"].map(self._convert_ids_to_categories)
        unique_categories = list(set([x["name"] for x in self.ontology]))
        for c in unique_categories:
            df = pd.concat(
                (
                    df,
                    df["categories"]
                    .apply(lambda x: 1 if c in x else 0)
                    .to_frame(name=c),
                ),
                axis=1,
            )
        return df

    @cached_property
    def ontology(self) -> Dict:
        """Reads AudioSet ontology."""
        with open(os.path.join(self.path, "ontology.json"), "r") as fp:
            return json.load(fp)

    def _load_df(self, relative_path: str) -> pd.DataFrame:
        df = pd.read_csv(
            os.path.join(self.path, f"{relative_path}.csv"),
            skiprows=2,
            sep=", ",
            engine="python",
        )
        df = self.map_to_classes(df)
        df["filename"] = df["filename"].apply(
            lambda x: os.path.join(relative_path, x)
        )
        orig_len = len(df)
        df = df.loc[
            df["filename"].apply(
                lambda x: os.path.isfile(
                    os.path.join(
                        self.features_path,
                        self.features_subdir,
                        x.replace(".wav", f".{self.file_type}"),
                    )
                )
            )
        ]
        avail_len = len(df)
        if avail_len != orig_len:
            self._log.warning(
                f"Found only {avail_len} "
                f"from the {orig_len} original samples "
                f"included in the {relative_path} set of Audioset. "
            )
        return df

    @cached_property
    def df_train(self) -> pd.DataFrame:
        df = self._load_df("balanced_train_segments")
        if self.use_unbalanced:
            u_df = self._load_df("unbalanced_train_segments")
            df = pd.concat((df, u_df))
        return df

    @cached_property
    def df_dev(self) -> pd.DataFrame:
        return self._load_df("eval_segments")

    @cached_property
    def df_test(self) -> pd.DataFrame:
        return self.df_dev

    @staticmethod
    def download(path: str) -> None:  # pragma: no cover
        """Download AudioSet.

        The data must be downloaded manually from
        https://research.google.com/audioset/download.html.
        We do not implement
        an automatic download
        as this is a very costly process
        that the user should handle
        outside of `autrainer`.

        Args:
            path: Path to the directory to download the dataset to.
        """
        pass



import argparse
from functools import wraps
import sys
from typing import Any, Callable, Dict, Optional, TypeVar

from .command_line_error import CommandLineError


F = TypeVar("F", bound=Callable[..., Any])
TQDM_LINE_ENDINGS = ["it/s]", "s/it]", "B/s]"]


def encode_override_kwargs(override_kwargs: Dict[str, Any]) -> Dict[str, str]:
    """Encode Hydra override kwargs by converting all values to strings.
    List values are converted to strings by joining the elements with commas.
    Boolean values are converted to lowercase strings.
    Duplicate keys are automatically overwritten.

    Args:
        override_kwargs: The override kwargs to encode.

    Returns:
        Encoded override kwargs.

    Raises:
        CommandLineError: If any override argument values are not of type list,
            str, int, float, or bool.
    """
    encoded_kwargs = {}
    for key, value in override_kwargs.items():
        if isinstance(value, list):
            value = ",".join(map(str, value))
        elif isinstance(value, bool):
            value = str(value).lower()
        if not isinstance(value, (str, int, float)):
            raise CommandLineError(
                f"Hydra override argument values must be of type list, "
                f"str, int, float, or bool, but got {type(value)}."
            )
        encoded_kwargs[key] = str(value)
    return encoded_kwargs


def run_hydra_cmd(
    cmd: str,
    override_kwargs: Optional[dict] = None,
    config_name: str = "config",
    config_path: Optional[str] = None,
    cmd_prefix: str = "autrainer",
) -> None:
    """Run a Hydra command in a subprocess and print the output.

    Args:
        cmd: The Hydra command to run.
        override_kwargs: Additional Hydra override arguments to pass to the
            train script.
        config_name: The name of the config (usually the file name without the
            .yaml extension). Defaults to "config".
        config_path: The config path, a directory where Hydra will search for
            config files. If config_path is None no directory is added to the
            search path. Defaults to None.
        cmd_prefix: The command prefix to use. Defaults to "autrainer".
    """
    import subprocess

    cmd = f"{cmd_prefix} {cmd} -cn {config_name}.yaml"
    if config_path is not None:
        cmd += f" -cp {config_path}"
    if override_kwargs is not None:
        for key, value in encode_override_kwargs(override_kwargs).items():
            cmd += f" {key}={value}"

    with subprocess.Popen(
        cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        shell=True,
        bufsize=1,
        encoding="utf-8",
        errors="replace",
        universal_newlines=False,
    ) as process:
        tqdm_line = False
        for line in process.stdout:
            line = line.rstrip()

            # TODO: find a better way to handle tqdm output
            if any(line.endswith(ending) for ending in TQDM_LINE_ENDINGS):
                print("\r" + line, end="")
                tqdm_line = True
            elif tqdm_line:
                print("\n" + line)
                tqdm_line = False
            else:
                print(line)

    process.stdout.close()
    process.wait()


def running_in_notebook() -> bool:
    return any("jupyter" in arg or "ipykernel" in arg for arg in sys.argv)


def catch_cli_errors(func: F) -> F:
    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        try:
            return func(*args, **kwargs)
        except CommandLineError as e:
            if not running_in_notebook():
                raise e
            print(e.message, file=sys.stderr)

    return wrapper


def add_hydra_args_to_sys(
    override_kwargs: Optional[dict] = None,
    config_name: str = "config",
    config_path: Optional[str] = None,
) -> None:
    if config_name.replace(".yaml", "") != "config":
        sys.argv.extend(["-cn", config_name])
    if config_path is not None:
        sys.argv.extend(["-cp", config_path])
    if override_kwargs is not None:
        sys.argv.extend(
            f"{k}={v}"
            for k, v in encode_override_kwargs(override_kwargs).items()
        )


def check_invalid_config_path_arg(parser: argparse.ArgumentParser) -> None:
    cp_parser = argparse.ArgumentParser()
    cp_parser.add_argument("--config-path", "-cp")
    args, remaining = cp_parser.parse_known_args()
    sys.argv = [sys.argv[0]] + remaining
    if args.config_path is None:
        return
    raise CommandLineError(
        parser=parser,
        message=(
            "The `--config-path/-cp` Hydra CLI argument is not supported "
            " by autrainer.\nTo override the config directory and name, "
            "use the `--config-dir/-cd` and `--config-name/-cn` arguments "
            "instead.\nFor more information, refer to the Hydra documentation "
            "(https://hydra.cc/docs/advanced/hydra-command-line-flags/) and "
            "https://github.com/autrainer/autrainer/pull/132."
        ),
    )



import os
import shutil
from typing import Dict, List, Optional, Union

from omegaconf import DictConfig
import pandas as pd

from autrainer.transforms import SmartCompose

from .abstract_dataset import BaseMLClassificationDataset
from .utils import ZipDownloadManager


FILES = {
    "EDANSA-2019.zip": "https://zenodo.org/records/6824272/files/EDANSA-2019.zip?download=1"
}


class EDANSA2019(BaseMLClassificationDataset):
    def __init__(
        self,
        path: str,
        features_subdir: Optional[str],
        seed: int,
        metrics: List[Union[str, DictConfig, Dict]],
        tracking_metric: Union[str, DictConfig, Dict],
        index_column: str,
        target_column: List[str],
        file_type: str,
        file_handler: Union[str, DictConfig, Dict],
        features_path: Optional[str] = None,
        train_transform: Optional[SmartCompose] = None,
        dev_transform: Optional[SmartCompose] = None,
        test_transform: Optional[SmartCompose] = None,
        stratify: Optional[List[str]] = None,
        threshold: float = 0.5,
    ) -> None:
        """EDANSA 2019 dataset.

        Args:
            path: Root path to the dataset.
            features_subdir: Subdirectory containing the features.
                If `None`, defaults to audio subdirectory,
                which is `default` for the standard format,
                but can be overridden in the dataset specification.
            seed: Seed for reproducibility.
            metrics: List of metrics to calculate.
            tracking_metric: Metric to track.
            index_column: Index column of the dataframe.
            target_column: Target column of the dataframe.
            file_type: File type of the features.
            file_handler: File handler to load the data.
            features_path: Root path to features. Useful
                when features need to be extracted and stored
                in a different directory than the root of the dataset.
                If `None`, will be set to `path`. Defaults to `None`.
            train_transform: Transform to apply to the training set.
                Defaults to None.
            dev_transform: Transform to apply to the development set.
                Defaults to None.
            test_transform: Transform to apply to the test set.
                Defaults to None.
            stratify: Columns to stratify the dataset on. Defaults to None.
            threshold: Threshold for classification. Defaults to 0.5.
        """
        super().__init__(
            path=path,
            features_subdir=features_subdir,
            seed=seed,
            metrics=metrics,
            tracking_metric=tracking_metric,
            index_column=index_column,
            target_column=target_column,
            file_type=file_type,
            file_handler=file_handler,
            features_path=features_path,
            train_transform=train_transform,
            dev_transform=dev_transform,
            test_transform=test_transform,
            stratify=stratify,
            threshold=threshold,
        )
        self._assert_target_column(allowed_columns=self.df_train.columns[9:])

    @staticmethod
    def download(path: str) -> None:  # pragma: no cover
        """Download the EDANSA 2019 dataset.

        For more information on the dataset, see:
        https://zenodo.org/doi/10.5281/zenodo.6824271

        Args:
            path: Path to the directory to download the dataset to.
        """
        out_path = os.path.join(path, "default")
        if os.path.isdir(out_path):
            return
        os.makedirs(out_path, exist_ok=True)

        # download and extract files
        dl_manager = ZipDownloadManager(FILES, path)
        dl_manager.download(check_exist=["EDANSA-2019"])
        dl_manager.extract(check_exist=["EDANSA-2019"])

        # move audio files
        for item in os.listdir(os.path.join(path, "EDANSA-2019", "data")):
            shutil.move(
                os.path.join(path, "EDANSA-2019", "data", item),
                os.path.join(out_path, item),
            )

        # process dataframes
        df = pd.read_csv(os.path.join(path, "EDANSA-2019", "labels.csv"))
        target_columns = df.columns[9:]
        for col in target_columns:
            df[col] = df[col].astype(float)
        df_train = df.loc[df["set"] == "train"]
        df_dev = df.loc[df["set"] == "valid"]
        df_test = df.loc[df["set"] == "test"]
        df_train.to_csv(os.path.join(path, "train.csv"), index=False)
        df_dev.to_csv(os.path.join(path, "dev.csv"), index=False)
        df_test.to_csv(os.path.join(path, "test.csv"), index=False)

        # remove unnecessary files
        shutil.rmtree(os.path.join(path, "EDANSA-2019"))



from contextlib import contextmanager
import io
import sys
from typing import Generator


@contextmanager
def silence() -> Generator[None, None, None]:
    """Context manager to suppress stdout and stderr.

    Yields:
        None.
    """
    new_stdout, new_stderr = io.StringIO(), io.StringIO()
    old_stdout, old_stderr = sys.stdout, sys.stderr
    try:
        sys.stdout, sys.stderr = new_stdout, new_stderr
        yield
    finally:
        sys.stdout, sys.stderr = old_stdout, old_stderr



import warnings

import torch
from transformers import Wav2Vec2Config, Wav2Vec2Model

from .abstract_model import AbstractModel
from .ffnn import FFNN


class W2V2Backbone(AbstractModel):
    def __init__(
        self,
        model_name,
        freeze_extractor: bool = True,
        time_pooling: bool = True,
        transfer: bool = False,
    ) -> None:
        self.model_name = model_name
        self.freeze_extractor = freeze_extractor
        self.time_pooling = time_pooling
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            if transfer:  # pragma: no cover
                model = Wav2Vec2Model.from_pretrained(self.model_name)
            else:
                config = Wav2Vec2Config.from_pretrained(
                    self.model_name,
                    output_hidden_states=True,
                    return_dict=True,
                )
                model = Wav2Vec2Model(config)
        super().__init__(
            output_dim=model.config.hidden_size,
            transfer=transfer,
        )
        self.model = model
        if self.freeze_extractor:
            self.model.freeze_feature_encoder()

    def embeddings(self, features: torch.Tensor) -> torch.Tensor:
        features = self.model(features)["last_hidden_state"]
        if self.time_pooling:
            features = features.mean(1)
        return features

    def forward(self, features: torch.Tensor) -> torch.Tensor:
        return self.embeddings(features)


class W2V2FFNN(AbstractModel):
    def __init__(
        self,
        output_dim: int,
        model_name: str,
        freeze_extractor: bool,
        hidden_size: int,
        num_layers: int = 2,
        dropout: float = 0.5,
        transfer: bool = False,
    ) -> None:
        """Wav2Vec2 model with FFNN frontend adapted for audio classification.
        For more information, see:
        https://huggingface.co/docs/transformers/model_doc/wav2vec2

        Args:
            output_dim: Output dimension of the FFNN.
            model_name: Name of the model loaded from Huggingface.
            freeze_extractor: Whether to freeze the feature extractor.
            hidden_size: Hidden size of the FFNN.
            num_layers: Number of layers of the FFNN. Defaults to 2.
            dropout: Dropout rate. Defaults to 0.5.
            transfer: Whether to initialize the Wav2Vec2 backbone with
                pretrained weights. Defaults to False.
        """
        super().__init__(output_dim, transfer)
        self.model_name = model_name
        self.freeze_extractor = freeze_extractor
        self.hidden_size = hidden_size
        self.num_layers = num_layers
        self.dropout = dropout
        self.backbone = W2V2Backbone(
            model_name=model_name,
            freeze_extractor=freeze_extractor,
            time_pooling=True,
            transfer=transfer,
        )
        self.frontend = FFNN(
            input_size=self.backbone.output_dim,
            hidden_size=hidden_size,
            output_dim=output_dim,
            num_layers=num_layers,
            dropout=dropout,
        )

    def embeddings(self, features: torch.Tensor) -> torch.Tensor:
        return self.backbone(features.squeeze(1))

    def forward(self, features: torch.Tensor) -> torch.Tensor:
        return self.frontend(self.embeddings(features))



from .abstract_target_transform import AbstractTargetTransform
from .label_encoder import LabelEncoder
from .min_max_scaler import MinMaxScaler
from .multi_label_encoder import MultiLabelEncoder
from .multi_target_min_max_scaler import MultiTargetMinMaxScaler


__all__ = [
    "AbstractTargetTransform",
    "LabelEncoder",
    "MinMaxScaler",
    "MultiLabelEncoder",
    "MultiTargetMinMaxScaler",
]



from typing import List

from .abstract_constants import AbstractConstants


class NamingConstants(AbstractConstants):
    """Singleton for managing the naming configurations of `autrainer`."""

    _name = "NamingConstants"
    _naming_convention = [
        "dataset",
        "model",
        "optimizer",
        "learning_rate",
        "batch_size",
        "training_type",
        "iterations",
        "scheduler",
        "augmentation",
        "seed",
    ]
    _invalid_aggregations = ["training_type"]
    _valid_aggregations = list(
        set(_naming_convention) - set(_invalid_aggregations)
    )
    _config_dirs = [
        "augmentation",
        "dataset",
        "model",
        "optimizer",
        "plotting",
        "preprocessing",
        "scheduler",
    ]

    @property
    def NAMING_CONVENTION(self) -> List[str]:
        """Get the naming convention of runs.
        Defaults to :attr:`["dataset", "model", "optimizer", "learning_rate",
        "batch_size", "training_type", "iterations", "scheduler",
        "augmentation", "seed"]`.

        Returns:
            Naming convention of runs.
        """
        return self._naming_convention

    @NAMING_CONVENTION.setter
    def NAMING_CONVENTION(self, naming_convention: List[str]) -> None:
        """Set the naming convention of runs.

        Args:
            naming_convention: Naming convention of runs.

        Raises:
            ValueError: If the naming convention is not a list of strings.
        """
        self._assert_type(naming_convention, list, "NAMING_CONVENTION")
        for n in naming_convention:
            self._assert_type(
                n,
                str,
                "NAMING_CONVENTION",
                "in naming convention",
            )
        self._naming_convention = naming_convention

    @property
    def INVALID_AGGREGATIONS(self) -> List[str]:
        """Get the invalid aggregations for postprocessing.
        Defaults to :attr:`["training_type"]`.

        Returns:
            Invalid aggregations for postprocessing.
        """
        return self._invalid_aggregations

    @INVALID_AGGREGATIONS.setter
    def INVALID_AGGREGATIONS(self, invalid_aggregations: List[str]) -> None:
        """Set the invalid aggregations for postprocessing.

        Args:
            invalid_aggregations: Invalid aggregations for postprocessing.

        Raises:
            ValueError: If the invalid aggregations are not a list of strings.
        """
        self._assert_type(invalid_aggregations, list, "INVALID_AGGREGATIONS")
        for i in invalid_aggregations:
            self._assert_type(
                i,
                str,
                "INVALID_AGGREGATIONS",
                "in invalid aggregations",
            )
        self._invalid_aggregations = invalid_aggregations

    @property
    def VALID_AGGREGATIONS(self) -> List[str]:
        """Get the valid aggregations for postprocessing.
        Defaults to :attr:`["dataset", "model", "optimizer", "learning_rate",
        "batch_size", "iterations", "scheduler", "augmentation", "seed"]`
        (the naming convention without the invalid aggregations).

        Returns:
            Valid aggregations for postprocessing.
        """
        return self._valid_aggregations

    @VALID_AGGREGATIONS.setter
    def VALID_AGGREGATIONS(self, valid_aggregations: List[str]) -> None:
        """Set the valid aggregations for postprocessing.

        Args:
            valid_aggregations: Valid aggregations for postprocessing.

        Raises:
            ValueError: If the valid aggregations are not a list of strings.
        """
        self._assert_type(valid_aggregations, list, "VALID_AGGREGATIONS")
        for v in valid_aggregations:
            self._assert_type(
                v,
                str,
                "VALID_AGGREGATIONS",
                "in valid aggregations",
            )
        self._valid_aggregations = valid_aggregations

    @property
    def CONFIG_DIRS(self) -> List[str]:
        """Get the configuration directories for Hydra configurations.
        Defaults to :attr:`["augmentation", "dataset", "model", "optimizer",
        "plotting", "preprocessing", "scheduler"]`.

        Returns:
            Configuration directories for Hydra configurations.
        """
        return self._config_dirs

    @CONFIG_DIRS.setter
    def CONFIG_DIRS(self, config_dirs: List[str]) -> None:
        """Set the configuration directories for Hydra configurations.

        Args:
            config_dirs: Configuration directories for Hydra configurations.

        Raises:
            ValueError: If the configuration directories are not a list of
                strings.
        """
        self._assert_type(config_dirs, list, "CONFIG_DIRS")
        for c in config_dirs:
            self._assert_type(
                c,
                str,
                "CONFIG_DIRS",
                "in configuration directories",
            )
        self._config_dirs = config_dirs



from .classification import (
    BalancedBCEWithLogitsLoss,
    BalancedCrossEntropyLoss,
    BCEWithLogitsLoss,
    CrossEntropyLoss,
    WeightedBCEWithLogitsLoss,
    WeightedCrossEntropyLoss,
)
from .regression import MSELoss, WeightedMSELoss


__all__ = [
    "BalancedBCEWithLogitsLoss",
    "BalancedCrossEntropyLoss",
    "BCEWithLogitsLoss",
    "CrossEntropyLoss",
    "MSELoss",
    "WeightedBCEWithLogitsLoss",
    "WeightedCrossEntropyLoss",
    "WeightedMSELoss",
]



from __future__ import annotations

import os
import shutil
from typing import TYPE_CHECKING

import pandas as pd

from autrainer.core.constants import NamingConstants
from autrainer.postprocessing.postprocessing_utils import (
    get_run_names,
    load_yaml,
)


if TYPE_CHECKING:
    from .training import ModularTaskTrainer  # pragma: no cover


class ContinueTraining:
    def __init__(self, run_name: str, remove_continued_runs: bool = False):
        self.run_name = run_name
        self.continued_run = None
        self.remove_continued_runs = remove_continued_runs

    def cb_on_train_begin(self, trainer: "ModularTaskTrainer") -> None:
        finished_runs = get_run_names(trainer.output_directory.parent)

        current_cfg = self._create_run_config(self.run_name)
        current_iteration = int(current_cfg.pop("iterations"))

        matches = {}
        for run in finished_runs:
            metrics = os.path.join(
                trainer.output_directory.parent, run, "metrics.csv"
            )
            if not os.path.exists(metrics):
                continue

            run_cfg = self._create_run_config(run)
            run_iteration = int(run_cfg.pop("iterations"))
            if current_cfg == run_cfg and current_iteration > run_iteration:
                matches[run] = current_iteration
        if not matches:
            return
        self.continue_training(trainer, max(matches, key=matches.get))

    def cb_on_train_end(self, trainer: "ModularTaskTrainer") -> None:
        if self.continued_run is None or not self.remove_continued_runs:
            return
        shutil.rmtree(
            os.path.join(trainer.output_directory.parent, self.continued_run)
        )

    def continue_training(self, trainer: ModularTaskTrainer, run: str) -> None:
        self.continued_run = run
        self._copy_dirs(trainer, run)
        self._init_metrics(trainer, run)
        self._init_timers(trainer, run)
        self._init_states(trainer)
        self._replay_loggers(trainer)

    @staticmethod
    def _create_run_config(run: str) -> dict:
        run_values = run.split("_")
        return {
            n: v
            for n, v in zip(NamingConstants().NAMING_CONVENTION, run_values)
        }

    def _copy_dirs(self, trainer: ModularTaskTrainer, run: str) -> None:
        dirs = ["_best", "_initial"]
        for d in dirs:
            shutil.rmtree(os.path.join(trainer.output_directory, d))
        iteration_folders = [
            f
            for f in os.listdir(
                os.path.join(trainer.output_directory.parent, run)
            )
            if f.startswith(trainer.cfg.training_type)
        ]
        dirs += iteration_folders
        for dir in dirs:
            shutil.copytree(
                os.path.join(trainer.output_directory.parent, run, dir),
                os.path.join(trainer.output_directory, dir),
            )

    def _init_metrics(self, trainer: ModularTaskTrainer, run: str) -> None:
        trainer.metrics = pd.read_csv(
            os.path.join(trainer.output_directory.parent, run, "metrics.csv"),
            index_col="iteration",
        )
        trainer.initial_iteration = int(trainer.metrics.index.max() + 1)
        m = trainer.data.tracking_metric
        trainer.max_dev_metric = m.get_best(trainer.metrics[m.name])
        trainer.best_iteration = m.get_best_pos(trainer.metrics[m.name])

    def _init_timers(self, trainer: ModularTaskTrainer, run: str) -> None:
        timers = load_yaml(
            os.path.join(trainer.output_directory.parent, run, "timer.yaml")
        )
        trainer.train_timer.time_log.append(timers["train"]["total_seconds"])
        trainer.dev_timer.time_log.append(timers["dev"]["total_seconds"])
        trainer.test_timer.time_log.append(timers["test"]["total_seconds"])

    def _init_states(self, trainer: ModularTaskTrainer) -> None:
        last_iteration = int(trainer.metrics.index.max())
        last_dir = os.path.join(
            trainer.output_directory,
            f"{trainer.cfg.training_type}_{last_iteration}",
        )
        trainer.bookkeeping.load_state(trainer.model, "model.pt", last_dir)
        trainer.model = trainer.model.to(trainer.DEVICE)
        trainer.bookkeeping.load_state(
            trainer.optimizer, "optimizer.pt", last_dir
        )
        if trainer.scheduler is not None:
            trainer.bookkeeping.load_state(
                trainer.scheduler, "scheduler.pt", last_dir
            )

    def _replay_loggers(self, trainer: ModularTaskTrainer) -> None:
        for logger in trainer.loggers:
            for iteration in trainer.metrics.index:
                logger.log_and_update_metrics(
                    trainer.metrics.loc[iteration].to_dict(), iteration
                )



import os
import warnings

import torch

from .abstract_model import AbstractModel
from .ffnn import FFNN


class TDNNFFNN(AbstractModel):
    def __init__(
        self,
        output_dim: int,
        hidden_size: int,
        num_layers: int = 2,
        dropout: float = 0.5,
        transfer: bool = False,
    ) -> None:
        """Time Delay Neural Network with FFNN frontend.

        Args:
            output_dim: Output dimension.
            hidden_size: Hidden size.
            num_layers: Number of layers. Defaults to 2.
            dropout: Dropout rate. Defaults to 0.5.
            transfer: Whether to initialize the TDNN backbone with
                pretrained weights. Defaults to False.
        """
        super().__init__(output_dim, transfer)  # no transfer learning weights
        self.hidden_size = hidden_size
        self.num_layers = num_layers
        self.dropout = dropout
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=FutureWarning)
            if transfer:  # pragma: no cover
                checkpoint_dir = os.path.join(
                    torch.hub.get_dir(), "speechbrain"
                )
                os.makedirs(checkpoint_dir, exist_ok=True)
                from speechbrain.inference.classifiers import EncoderClassifier

                tdnn = EncoderClassifier.from_hparams(
                    source="speechbrain/spkrec-ecapa-voxceleb",
                    freeze_params=False,
                    savedir=checkpoint_dir,
                )
                self.features = tdnn.mods["compute_features"]
                self.backbone = tdnn.mods["embedding_model"]
            else:
                from speechbrain.lobes.features import Fbank
                from speechbrain.lobes.models.ECAPA_TDNN import ECAPA_TDNN

                # taken from https://huggingface.co/speechbrain/spkrec-ecapa-voxceleb/blob/main/hyperparams.yaml
                self.features = Fbank(n_mels=80)
                self.backbone = ECAPA_TDNN(
                    input_size=80,
                    channels=[1024, 1024, 1024, 1024, 3072],
                )

        self.frontend = FFNN(
            input_size=list(self.backbone.modules())[-1].weight.shape[0],
            hidden_size=hidden_size,
            output_dim=output_dim,
            num_layers=num_layers,
            dropout=dropout,
        )

    def embeddings(self, features: torch.Tensor) -> torch.Tensor:
        _device = features.device
        features = self.features(features.squeeze(1).cpu())
        features = features.to(_device)
        embs = self.backbone(features).squeeze(1)
        return embs

    def forward(self, features: torch.Tensor) -> torch.Tensor:
        embs = self.embeddings(features)
        return self.frontend(embs)



from .abstract_transform import AbstractTransform
from .smart_compose import SmartCompose
from .specific_transforms import (
    AnyToTensor,
    Expand,
    FeatureExtractor,
    GrayscaleToRGB,
    ImageToFloat,
    Normalize,
    NumpyToTensor,
    OpenSMILE,
    PannMel,
    RandomCrop,
    Resample,
    Resize,
    RGBAToRGB,
    RGBToGrayscale,
    ScaleRange,
    SpectToImage,
    SquarePadCrop,
    Standardizer,
    StereoToMono,
)
from .transform_manager import TransformManager


__all__ = [
    "AbstractTransform",
    "AnyToTensor",
    "Expand",
    "FeatureExtractor",
    "GrayscaleToRGB",
    "ImageToFloat",
    "Normalize",
    "NumpyToTensor",
    "OpenSMILE",
    "PannMel",
    "RandomCrop",
    "Resample",
    "Resize",
    "RGBAToRGB",
    "RGBToGrayscale",
    "ScaleRange",
    "SmartCompose",
    "SpectToImage",
    "SquarePadCrop",
    "StereoToMono",
    "Standardizer",
    "TransformManager",
]



from typing import Any, Dict, List, Union

import autrainer
from autrainer.transforms import SmartCompose

from .abstract_augmentation import AbstractAugmentation
from .utils import assign_seeds, convert_shorthand


class AugmentationPipeline:
    def __init__(
        self,
        pipeline: List[Union[str, Dict[str, Any]]],
        generator_seed: int = 0,
        increment: bool = True,
    ) -> None:
        """Initialize an augmentation pipeline.

        Args:
            pipeline: The list of augmentations to apply.
            generator_seed: Seed to pass to each augmentation for
                reproducibility if the augmentation does not have a seed.
                Defaults to 0.
            increment: Whether to increment the generator seed for each
                augmentation that does not define its own seed.
                Defaults to True.
        """
        self.generator_seed = generator_seed
        self.pipeline = []
        current_seed = generator_seed
        for aug in pipeline:
            aug = convert_shorthand(aug)
            aug, current_seed = assign_seeds(aug, current_seed, increment)
            self.pipeline.append(aug)

        self.pipeline: List[AbstractAugmentation] = [
            autrainer.instantiate_shorthand(
                aug,
                instance_of=AbstractAugmentation,
            )
            for aug in self.pipeline
        ]

    def create_pipeline(self) -> SmartCompose:
        """Create the composed and ordered augmentation pipeline.

        Returns:
            Composed augmentation pipeline.
        """
        return SmartCompose(self.pipeline)



import os
import shutil
from typing import Dict, List, Optional, Tuple, Union

from omegaconf import DictConfig, ListConfig

from .aggregate import AggregateGrid
from .postprocessing_utils import load_yaml
from .summarize import SummarizeGrid


class GroupGrid:
    def __init__(
        self,
        results_dir: str,
        groupings: Union[ListConfig[DictConfig], List[Dict]],
        max_runs: Optional[int] = None,
        plot_params: Optional[dict] = None,
    ) -> None:
        """Group runs of one or more grid search experiments based on the
        specified groupings.

        Args:
            results_dir: The directory where the results are stored.
            groupings: A list of experiments to create containing one or more
                runs to group.
            max_runs_plot: The maximum number of best runs to plot. If None,
                all runs will be plotted. Defaults to None.
            plot_params: Additional parameters for plotting. Defaults to None.
        """
        self.results_dir = results_dir
        self.groupings = groupings
        self.max_runs = max_runs
        self.plot_params = plot_params

    def group_runs(self) -> None:
        """Group the runs of the specified experiments based on the groupings."""
        for grouping in self.groupings:
            grouping = self._preprocess(grouping)
            mappings = self._copy_runs(grouping.experiment_id, grouping.runs)
            self._group_experiment(
                grouping.experiment_id,
                grouping.runs,
                grouping.create_summary,
                mappings,
            )

    def _preprocess(self, grouping: DictConfig) -> DictConfig:
        set_dir = grouping.get("dir") is not None
        set_id = grouping.get("id") is not None
        set_states = grouping.get("states") is not None

        if not set_dir and not set_id and not set_states:
            return grouping

        for run in grouping.runs:
            if set_dir and run.get("dir") is None:
                run.dir = grouping.dir
            if set_id and run.get("id") is None:
                run.id = grouping.id
            if set_states and run.get("states") is None:
                run.states = grouping.states
        return grouping

    def _copy_runs(
        self, experiment: str, runs: ListConfig[DictConfig]
    ) -> Dict[str, List[str]]:
        run_paths, copy_states = self._resolve_runs_paths(experiment, runs)
        run_paths, copy_states, mappings = self._resolve_aggregated_runs(
            run_paths, copy_states
        )
        for run_path, copy_state in zip(run_paths, copy_states):
            self._copy_run(experiment, run_path, copy_state)
        return mappings

    def _resolve_runs_paths(
        self, experiment: str, runs: ListConfig[DictConfig]
    ) -> Tuple[List[str], List[bool]]:
        paths = []
        copy_states = []
        for run in runs:
            base_path = os.path.join(run.dir, run.id)
            dirs = [d for d in os.listdir(base_path) if d.startswith("agg_")]
            dirs.append("training")
            if isinstance(run.combine, str):
                run.combine = [run.combine]
            for r in run.combine:
                for d in dirs:
                    p = os.path.join(base_path, d, r)
                    if os.path.exists(os.path.join(p, "metrics.csv")):
                        paths.append(p)
                        copy_states.append(run.states)
                        break
                else:
                    raise ValueError(
                        f"Run '{r}' to combine to '{run.run_name}'"
                        f"in experiment '{experiment}' does not exist "
                        f"in experiment '{run.id}'"
                    )
        return paths, copy_states

    def _resolve_aggregated_runs(
        self, run_paths: List[str], copy_states: List[bool]
    ) -> Tuple[List[str], List[bool], Dict[str, List[str]]]:
        expanded_paths = []
        expanded_states = []
        mappings = {}
        for run_path, copy_state in zip(run_paths, copy_states):
            mappings[os.path.basename(run_path)] = []
            for run in self._get_aggregated_runs(run_path):
                path = os.path.join(
                    os.path.dirname(os.path.dirname(run_path)), "training", run
                )
                expanded_paths.append(path)
                expanded_states.append(copy_state)
                mappings[os.path.basename(run_path)].append(run)

        return expanded_paths, expanded_states, mappings

    def _get_aggregated_runs(self, run_path: str) -> List[str]:
        if "#" not in os.path.basename(run_path):
            return [os.path.basename(run_path)]
        return load_yaml(os.path.join(run_path, "runs.yaml"))

    def _copy_run(
        self, experiment: str, run_path: str, copy_state: bool
    ) -> None:
        dst = os.path.join(
            self.results_dir,
            experiment,
            "training",
            os.path.basename(run_path),
        )
        if os.path.exists(
            os.path.join(dst, "metrics.csv")
        ) or not os.path.exists(run_path):
            return
        ignore = None if copy_state else shutil.ignore_patterns("*.pt")
        shutil.copytree(
            src=run_path,
            dst=dst,
            ignore=ignore,
        )

    def _group_experiment(
        self,
        experiment: str,
        runs: ListConfig[DictConfig],
        create_summary: bool,
        mappings: Dict[str, List[str]],
    ) -> None:
        grouped_dict = {}
        for run in runs:
            base_runs = []
            for r in run.combine:
                base_runs.extend(mappings[r])

            grouped_dict.update({run.run_name: base_runs})
        if create_summary:
            sg = SummarizeGrid(
                results_dir=self.results_dir,
                experiment_id=experiment,
                max_runs_plot=self.max_runs,
            )
            sg.summarize()
            sg.plot_aggregated_bars()
            sg.plot_metrics()

        ag = AggregateGrid(
            results_dir=self.results_dir,
            experiment_id=experiment,
            aggregate_list="",
            aggregate_name="grouping",
            aggregated_dict=grouped_dict,
            max_runs_plot=self.max_runs,
            plot_params=self.plot_params,
        )
        ag.aggregate()
        ag.summarize()



from typing import Dict, List

import torch

from .abstract_target_transform import AbstractTargetTransform


class MultiTargetMinMaxScaler(AbstractTargetTransform):
    def __init__(
        self,
        target: List[str],
        minimum: List[float],
        maximum: List[float],
    ) -> None:
        """Minimum-Maximum Scaler for multi-target regression.

        Args:
            target: Names of the targets.
            minimum: Minimum values of all target values.
            maximum: Maximum values of all target values.

        Raises:
            ValueError: If minimum is not less than maximum.
        """
        for m, M in zip(minimum, maximum):
            if not m < M:
                raise ValueError(
                    f"Minimum '{m}' must be less than maximum '{M}'."
                )
        self.target = target
        self.minimum = [float(m) for m in minimum]
        self._m = torch.Tensor(self.minimum)
        self.maximum = [float(M) for M in maximum]
        self._M = torch.Tensor(self.maximum)

    def __len__(self) -> int:
        """Get the number of unique targets.

        Returns:
            Number of unique targets.
        """
        return len(self.target)

    def encode(self, x: List[float]) -> torch.Tensor:
        """Encode a target value by scaling it between the minimum and maximum.

        Args:
            x: Target value.

        Returns:
            Scaled target value.
        """
        return (torch.Tensor(x) - self._m) / (self._M - self._m)

    def decode(self, x: List[float]) -> List[float]:
        """Decode a target value by reversing the scaling between the minimum
        and maximum. Inverse operation of encode.

        Args:
            x: Scaled target value.

        Returns:
            Unscaled target value.
        """
        return (torch.Tensor(x) * (self._M - self._m) + self._m).tolist()

    def probabilities_training(self, x: torch.Tensor) -> torch.Tensor:
        """Get the encoded probabilities from a batch of model outputs
        during training by applying the sigmoid function.

        Args:
            x: Batch of model outputs.

        Returns:
            Encoded probabilities.
        """
        return torch.sigmoid(x)

    def probabilities_inference(self, x: torch.Tensor) -> torch.Tensor:
        """Get the encoded probabilities from a batch of model outputs by
        applying the sigmoid function.

        Args:
            x: Batch of model outputs.

        Returns:
            Encoded probabilities.
        """
        return torch.sigmoid(x)

    def predict_inference(self, x: torch.Tensor) -> List[List[float]]:
        """Get the encoded predictions from a batch of model output
        probabilities by returning the raw values.

        Args:
            x: Batch of model output probabilities.

        Returns:
            Encoded predictions.
        """
        return x.tolist()

    def majority_vote(self, x: List[List[float]]) -> List[float]:
        """Get the majority vote from a list of target values by averaging
        the predictions.

        Args:
            x: List of target values.

        Returns:
            Average target value.
        """
        return [sum(item) / len(item) for item in zip(*x)]

    def probabilities_to_dict(self, x: torch.Tensor) -> Dict[str, float]:
        """Convert a tensor of probabilities to a dictionary of targets and
        their probabilities.

        Args:
            x: Tensor of probabilities.

        Returns:
            Dictionary of targets and their probabilities.
        """
        return {label: prob.item() for label, prob in zip(self.target, x)}



from typing import TYPE_CHECKING, List, Optional
import warnings

from audtorch import transforms as AT
from matplotlib.pyplot import get_cmap
import numpy as np
from omegaconf import OmegaConf
from PIL import Image
import torch
from torchaudio import transforms as TT
import torchlibrosa
from torchvision import transforms as T
from transformers import (
    ASTFeatureExtractor,
    AutoFeatureExtractor,
    Wav2Vec2FeatureExtractor,
    WhisperFeatureExtractor,
)

from autrainer.core.structs import AbstractDataItem

from .abstract_transform import AbstractTransform
from .smart_compose import SmartCompose
from .utils import to_numpy, to_tensor


try:
    import opensmile

    OPENSMILE_AVAILABLE = True
except ImportError:  # pragma: no cover
    OPENSMILE_AVAILABLE = False

if TYPE_CHECKING:  # pragma: no cover
    from autrainer.datasets import AbstractDataset
    from autrainer.datasets.utils import DatasetWrapper


FE_MAPPINGS = {
    "AST": {"fe": ASTFeatureExtractor, "padding": "max_length"},
    "Whisper": {"fe": WhisperFeatureExtractor, "padding": "max_length"},
    "W2V2": {"fe": Wav2Vec2FeatureExtractor, "padding": "longest"},
    None: {"fe": AutoFeatureExtractor, "padding": "max_length"},
}


class AnyToTensor(AbstractTransform):
    def __init__(self, order: int = -100) -> None:
        """Convert a numpy array, torch tensor, or a PIL image to a torch
        tensor.

        Args:
            order: The order of the transform in the pipeline. Defaults to -100.
        """
        super().__init__(order=order)
        self._convert_pil = T.ToTensor()

    def __call__(self, item: AbstractDataItem) -> AbstractDataItem:
        if isinstance(item.features, torch.Tensor):
            return item
        if isinstance(item.features, np.ndarray):
            item.features = to_tensor(item.features)
        elif isinstance(item.features, Image.Image):
            item.features = self._convert_pil(item.features) * 255
            item.features = item.features.to(torch.uint8)
        else:
            raise TypeError(
                "Input must be a 'torch.Tensor', 'np.ndarray', or "
                "'PIL.Image.Image', but got "
                f"'{item.features.__class__.__name__}'"
            )
        return item


class NumpyToTensor(AbstractTransform):
    def __init__(self, order: int = -100) -> None:
        """Convert a numpy array to a torch tensor.

        Args:
            order: The order of the transform in the pipeline. Defaults to -100.
        """
        super().__init__(order=order)

    def __call__(self, item: AbstractDataItem) -> AbstractDataItem:
        item.features = to_tensor(item.features)
        return item


class SpectToImage(AbstractTransform):
    def __init__(
        self,
        height: int,
        width: int,
        cmap: str = "magma",
        order: int = -90,
    ):
        """Convert a spectrogram in the range [0, 1] to a 3-channel uint8 image
        in the range [0, 255] using a specific colormap.

        Note: The transform should be used in combination with
        `~autrainer.transforms.ImageToFloat` to convert the image back to the
        range [0, 1] for training.

        Args:
            height: The height of the image.
            width: The width of the image.
            cmap: The colormap to use. Defaults to "magma".
            order: The order of the transform in the pipeline. Defaults to -90.
        """
        super().__init__(order=order)
        self.height = height
        self.width = width
        self.cmap = cmap
        self._cmap = get_cmap(cmap)

    def __call__(self, item: AbstractDataItem) -> AbstractDataItem:
        if item.features.dtype == torch.uint8:
            item.features = item.features / 255
        im = item.features.numpy().squeeze()
        im -= im.min()
        im /= im.max()
        im = (
            Image.fromarray(np.uint8(self._cmap(im)[..., :3] * 255))
            .transpose(Image.FLIP_TOP_BOTTOM)
            .resize((self.width, self.height))
        )
        item.features = torch.from_numpy(np.array(im)).transpose(0, 2)
        return item


class PannMel(AbstractTransform):
    def __init__(
        self,
        window_size: int,
        hop_size: int,
        sample_rate: int,
        fmin: int,
        fmax: int,
        mel_bins: int,
        ref: float,
        amin: float,
        top_db: int,
        order: int = -90,
    ):
        """Create a log-Mel spectrogram from an audio signal analogous to
        the Pretrained Audio Neural Networks (PANN) implementation.

        For more information, see:
        https://doi.org/10.48550/arXiv.1912.10211

        Args:
            window_size: The size of the window.
            hop_size: Hop length.
            sample_rate: The sample rate of the audio signal.
            fmin: The minimum frequency.
            fmax: The maximum frequency.
            mel_bins: The number of mel bins.
            ref: The reference amplitude.
            amin: The minimum amplitude.
            top_db: The top decibel.
            order: The order of the transform in the pipeline. Defaults to -90.
        """
        self.window_size = window_size
        self.hop_size = hop_size
        self.sample_rate = sample_rate
        self.fmin = fmin
        self.fmax = fmax
        self.mel_bins = mel_bins
        self.ref = ref
        self.amin = amin
        self.top_db = top_db
        super().__init__(order=order)
        self._spectrogram = torchlibrosa.stft.Spectrogram(
            n_fft=self.window_size,
            win_length=self.window_size,
            hop_length=self.hop_size,
        )
        self._mel = torchlibrosa.stft.LogmelFilterBank(
            sr=self.sample_rate,
            fmin=self.fmin,
            fmax=self.fmax,
            n_mels=self.mel_bins,
            n_fft=self.window_size,
            ref=self.ref,
            amin=self.amin,
            top_db=self.top_db,
        )

    def __call__(self, item: AbstractDataItem) -> AbstractDataItem:
        item.features = self._mel(self._spectrogram(item.features)).squeeze(1)
        return item


class Resize(AbstractTransform):
    def __init__(
        self,
        height: int = 64,
        width: int = 64,
        interpolation: str = "bilinear",
        antialias: bool = True,
        order: int = -90,
    ) -> None:
        """Resize an image to a specific height and width.
        If "Any" is provided for the height or width, the other dimension will
        be used as the size, creating a square image.

        Args:
            height: The target height. If set to "Any", the width will be used
                as the size. Defaults to 64.
            width: The target width. If set to "Any", the height will be used
                as the size. Defaults to 64.
            interpolation: The interpolation method to use. Defaults to
                'bilinear'.
            antialias: Whether to use antialiasing. Defaults to True.
            order: The order of the transform in the pipeline. Defaults to -90.
        """
        super().__init__(order=order)
        self.width = width
        self.height = height
        self.antialias = antialias
        self.interpolation = interpolation
        if self.width == "Any" and self.height == "Any":
            raise ValueError(
                "At least one of 'height' or 'width' must be specified."
            )
        if self.height == "Any":
            self._size = self.width
        elif self.width == "Any":
            self._size = self.height
        else:
            self._size = (self.height, self.width)
        self._resize = T.Resize(
            size=self._size,
            interpolation=getattr(
                T.InterpolationMode, self.interpolation.upper()
            ),
            antialias=self.antialias,
        )

    def __call__(self, item: AbstractDataItem) -> AbstractDataItem:
        item.features = self._resize(item.features)
        return item


class SquarePadCrop(AbstractTransform):
    def __init__(self, mode: str = "pad", order: int = -91) -> None:
        """Pad or crop an image to make it square.
        If the image is padded, the padding will be added to the shorter
        sides as black pixels.

        Args:
            mode: The mode to use in ["pad", "crop"]. Defaults to "pad".
            order: The order of the transform in the pipeline. Defaults to -91.

        Raises:
            ValueError: If an invalid mode is provided.
        """
        super().__init__(order=order)
        self.mode = mode
        if self.mode == "pad":
            self._mode_fn = max
        elif self.mode == "crop":
            self._mode_fn = min
        else:
            raise ValueError(f"Invalid mode '{self.mode}'.")

    def __call__(self, item: AbstractDataItem) -> AbstractDataItem:
        m = self._mode_fn(item.features.shape[-2:])
        item.features = T.functional.center_crop(item.features, m)
        return item


class ScaleRange(AbstractTransform):
    def __init__(
        self,
        range: List[float] = [0.0, 1.0],
        order: int = 90,
    ) -> None:
        """Scale a tensor to a specific target range.

        Args:
            range: The range to scale the tensor to. Defaults to [0.0, 1.0].
            order: The order of the transform in the pipeline. Defaults to 90.
        """
        super().__init__(order=order)
        if len(range) != 2:
            raise ValueError(
                f"Expected 'range' to be a list of length 2 but got {range}."
            )
        if range[0] == range[1]:
            raise ValueError(
                "The lower bound of 'range' must be different from the upper "
                f"bound but got {range}."
            )
        self.range = sorted(range)

    def __call__(self, item: AbstractDataItem) -> AbstractDataItem:
        m, M, (r0, r1) = item.features.min(), item.features.max(), self.range
        if m == M:
            item.features = torch.full_like(item.features, m)
            return item
        item.features = (item.features - m) / (M - m)
        item.features = item.features * (r1 - r0) + r0
        return item


class ImageToFloat(AbstractTransform):
    def __init__(self, order: int = 90) -> None:
        """Transform a uint8 image in the range [0, 255] to a float32 image in
        the range [0.0, 1.0].

        Args:
            order: The order of the transform in the pipeline. Defaults to 90.
        """
        super().__init__(order=order)

    def __call__(self, item: AbstractDataItem) -> AbstractDataItem:
        if item.features.dtype == torch.uint8:
            item.features = item.features.float() / 255
        return item


class Normalize(AbstractTransform):
    def __init__(
        self,
        mean: List[float],
        std: List[float],
        order: int = 95,
    ) -> None:
        """Normalize a tensor with a mean and standard deviation.

        Args:
            mean: The mean to use for normalization.
            std: The standard deviation to use for normalization.
            order: The order of the transform in the pipeline. Defaults to 95.
        """
        super().__init__(order=order)
        self.mean = mean
        self.std = std
        self._mean = torch.as_tensor(mean)
        self._std = torch.as_tensor(std)

    def __call__(self, item: AbstractDataItem) -> AbstractDataItem:
        # reshapes mean and std to match the data dimensions and broadcast
        # the values along the correct axes
        views = {
            3: (-1, 1, 1),  # 3D images: normalize along the channel axis
            2: (-1, 1),  # 2D audio: normalize along the channel axis
            1: (-1,),  # 1D tabular data: normalize along the feature axis
        }

        for dim, axes in views.items():
            if item.features.ndim == dim:
                break
        else:
            raise ValueError(
                f"Unsupported feature dimensions: {item.features.shape}"
            )

        mean, std = self._mean.view(*axes), self._std.view(*axes)
        item.features = item.features.to(torch.float32).sub(mean).div(std)
        return item


class Standardizer(AbstractTransform):
    def __init__(
        self,
        mean: Optional[List[float]] = None,
        std: Optional[List[float]] = None,
        subset: str = "train",
        order: int = -99,
    ) -> None:
        """Standardize a dataset by calculating the mean and standard deviation
        of the data and applying the normalization.

        The mean and standard deviation are automatically calculated from the
        data in the specified subset.

        Note: The transform is applied at the specified order in the pipeline
        such that any preceding transforms are applied before the mean and
        standard deviation are calculated.

        Args:
            mean: The mean to use for normalization. If None, the mean will be
                calculated from the data. Defaults to None.
            std: The standard deviation to use for normalization. If None, the
                standard deviation will be calculated from the data. Defaults
                to None.
            subset: The subset to use for calculating the mean and standard
                deviation. Must be one of ["train", "dev", "test"]. Defaults to
                "train".
            order: The order of the transform in the pipeline. Defaults to -99.
        """
        super().__init__(order=order)
        self._validate_subset(subset)
        self.subset = subset
        self.mean = mean
        self.std = std
        self._transform = self._init_transform()

    def _validate_subset(self, subset: str) -> None:
        if subset not in {"train", "dev", "test"}:
            raise ValueError(
                f"Invalid subset '{subset}'. Must be one of 'train', 'dev', "
                "or 'test'."
            )

    def _init_transform(self) -> Optional[AbstractTransform]:
        if self.mean and self.std:
            return Normalize(self.mean, self.std, self.order)

    def _collect_data(self, data: "AbstractDataset") -> torch.Tensor:
        ds: DatasetWrapper = getattr(data, f"{self.subset}_dataset")
        original_transform = ds.transform

        idx = original_transform.transforms.index(self)  # thanks to audobject
        preceding = SmartCompose(original_transform.transforms[:idx])
        ds.transform = preceding

        data = torch.stack([x.features for x in ds]).to(torch.float32)
        ds.transform = original_transform
        return data

    def setup(self, data: "AbstractDataset") -> None:
        if self.mean and self.std:  # if already computed for the current data
            return

        # select all dimensions except the dimension along which to normalize
        views = {
            4: (0, 2, 3),  # 3D images: normalize along the channel axis
            3: (0, 2),  # 2D audio: normalize along the channel axis
            2: (0,),  # 1D tabular data: normalize along the feature axis
        }
        collected = self._collect_data(data)
        for dim, axes in views.items():
            if collected.ndim == dim:
                break
        else:
            raise ValueError(f"Unsupported data dimensions: {collected.shape}")

        self.mean = collected.mean(dim=axes).tolist()
        self.std = collected.std(dim=axes).tolist()
        self._transform = self._init_transform()

    def __call__(self, item: AbstractDataItem) -> AbstractDataItem:
        if self._transform is None:
            raise ValueError("Standardize transform not initialized.")
        return self._transform(item)


class FeatureExtractor(AbstractTransform):
    def __init__(
        self,
        fe_type: Optional[str] = None,
        fe_transfer: Optional[str] = None,
        sampling_rate: int = 16000,
        order: int = -80,
    ) -> None:
        """Extract features from an audio signal using a feature extractor
        from the Hugging Face Transformers library.

        Args:
            fe_type: The class of feature extractor to use in ["AST", "Whisper",
                "W2V2", None]. If None, the AutoFeatureExtractor will be used.
                Defaults to None.
            fe_transfer: The name of a pretrained feature extractor to use.
                If None, the feature extractor will be initialized with default
                values. Defaults to None.
            sampling_rate: The sampling rate of the audio signal. Defaults to
                16000.
            order: The order of the transform in the pipeline. Defaults to -80.

        Raises:
            ValueError: If neither 'fe_type' nor 'fe_transfer' is provided.
        """
        super().__init__(order=order)
        if fe_type is None and fe_transfer is None:
            raise ValueError(
                "Either 'fe_type' or 'fe_transfer' must be provided."
            )
        self.fe_type = fe_type
        self.fe_transfer = fe_transfer
        self.sampling_rate = sampling_rate
        fe_class = FE_MAPPINGS[self.fe_type]["fe"]
        padding = FE_MAPPINGS[self.fe_type]["padding"]

        if self.fe_transfer is not None:
            feature_extractor = fe_class.from_pretrained(self.fe_transfer)
        else:
            feature_extractor = fe_class()
            extractor_dict = {
                k: repr(v) for k, v in feature_extractor.__dict__.items()
            }
            warnings.warn(
                f"{fe_class.__name__} "
                "initialized with default values:\n"
                f"{OmegaConf.to_yaml(extractor_dict)}"
            )

        def extract_features(signal: np.ndarray) -> torch.Tensor:
            if len(signal.shape) == 2:
                signal = signal.mean(0)
            extracted = feature_extractor(
                signal,
                sampling_rate=self.sampling_rate,
                padding=padding,
                return_tensors="pt",
            )
            return extracted[list(extracted.keys())[0]][0]

        self._extract_features = extract_features

    def __call__(self, item: AbstractDataItem) -> AbstractDataItem:
        item.features = self._extract_features(item.features.numpy())
        return item


class Expand(AbstractTransform):
    def __init__(
        self,
        size: int,
        method: str = "pad",
        axis: int = -2,
        order: int = -85,
    ) -> None:
        """Expand a tensor along a specific axis to a specific size, ensuring
        it is at least the specified size.
        If the tensor is smaller than the target size, it will be padded
        with zeros.
        If the tensor is larger than the target size, it will not be cropped.

        Args:
            size: The target size.
            method: The method to use in ["pad", "replicate"]. Defaults to
                "pad".
            axis: The axis along which to expand. Defaults to -2.
            order: The order of the transform in the pipeline. Defaults to -85.
        """
        super().__init__(order=order)
        self.size = size
        self.method = method
        self.axis = axis
        self._expand = AT.Expand(
            size=self.size, method=self.method, axis=self.axis
        )
        self._to_tensor = AnyToTensor()

    def __call__(self, item: AbstractDataItem) -> AbstractDataItem:
        item.features = self._expand(item.features)
        return self._to_tensor(item)


class RandomCrop(AbstractTransform):
    def __init__(
        self,
        size: int,
        method: str = "pad",
        axis: int = -2,
        fix_randomization: bool = False,
        order: int = -90,
    ) -> None:
        """Randomly crop a tensor along a specific axis to a specific size,
        ensuring it is the specified size.

        If the tensor is larger than the target size, it will be randomly
        cropped along the specified axis.

        If the tensor is smaller than the target size, it will be padded.

        Args:
            size: The target size.
            method: Padding method to use if the tensor is smaller than the
                target size in ["pad", "replicate"]. Defaults to "pad".
            axis: The axis along which to crop. Defaults to -2.
            fix_randomization: Whether to fix the randomization. Defaults to
                False.
            order: The order of the transform in the pipeline. Defaults to -90.
        """
        super().__init__(order=order)
        self.size = size
        self.method = method
        self.axis = axis
        self.fix_randomization = fix_randomization
        self._random_crop = AT.RandomCrop(
            size=self.size,
            method=self.method,
            axis=self.axis,
            fix_randomization=self.fix_randomization,
        )
        self._to_tensor = AnyToTensor()

    def __call__(self, item: AbstractDataItem) -> AbstractDataItem:
        item.features = self._random_crop(item.features)
        return self._to_tensor(item)


class StereoToMono(AbstractTransform):
    def __init__(self, order: int = -95) -> None:
        """Convert a stereo audio signal to mono by taking the mean of the
        first dimension.

        Args:
            order: The order of the transform in the pipeline. Defaults to -95.
        """
        super().__init__(order=order)

    def __call__(self, item: AbstractDataItem) -> AbstractDataItem:
        item.features = item.features.mean(0, keepdim=True)
        return item


class RGBAToRGB(AbstractTransform):
    """Convert an RGBA image to RGB by dropping the alpha channel.

    Args:
        order: The order of the transform in the pipeline. Defaults to -95.
    """

    def __init__(self, order: int = -95) -> None:
        super().__init__(order=order)

    def __call__(self, item: AbstractDataItem) -> AbstractDataItem:
        item.features = item.features[:3]
        return item


class RGBToGrayscale(AbstractTransform):
    def __init__(self, order: int = 100) -> None:
        """Convert an RGB or RGBA image to a grayscale image.

        For the conversion to grayscale, the
        luminance is calculated in line with the implementation in
        torchvision.transforms.Grayscale:
        Y = 0.2989 * R + 0.587 * G + 0.114 * B.

        Args:
            order: The order of the transform in the pipeline. Defaults to 100.
        """
        super().__init__(order=order)
        self._to_grayscale = T.Grayscale()

    def __call__(self, item: AbstractDataItem) -> AbstractDataItem:
        item.features = self._to_grayscale(item.features[:3])
        return item


class GrayscaleToRGB(AbstractTransform):
    def __init__(self, order: int = 100) -> None:
        """Convert a grayscale image to an RGB image.

        Args:
            order: The order of the transform in the pipeline. Defaults to 100.
        """
        super().__init__(order=order)

    def __call__(self, item: AbstractDataItem) -> AbstractDataItem:
        item.features = to_tensor(AT.Upmix(3, axis=0)(to_numpy(item.features)))
        return item


class Resample(AbstractTransform):
    def __init__(
        self,
        current_sr: int,
        target_sr: int,
        order: int = -95,
    ) -> None:
        """Resample an audio signal to a target sample rate.

        Args:
            current_sr: The current sample rate.
            target_sr: The target sample rate.
            order: The order of the transform in the pipeline. Defaults to -95.
        """
        self.current_sr = current_sr
        self.target_sr = target_sr
        super().__init__(order=order)
        self._resample = TT.Resample(
            orig_freq=self.current_sr, new_freq=self.target_sr
        )

    def __call__(self, item: AbstractDataItem) -> AbstractDataItem:
        item.features = self._resample(item.features)
        return item


class OpenSMILE(AbstractTransform):
    def __init__(
        self,
        feature_set: str,
        sample_rate: int,
        functionals: bool = False,
        order: int = -80,
    ) -> None:
        """Extract features from an audio signal using openSMILE.

        The openSMILE library must be installed to use this transform.
        To install the required extras, run:
        'pip install autrainer[opensmile]'.

        Args:
            feature_set: The feature set to use.
            sample_rate: The sample rate of the audio signal.
            functionals: Whether to use functionals. Defaults to False.
            order: The order of the transform in the pipeline. Defaults to -80.

        Raises:
            ImportError: If openSMILE is not available.
        """
        if not OPENSMILE_AVAILABLE:
            raise ImportError(
                "openSMILE is not available for feature extraction. Install "
                "the required extras with 'pip install autrainer[opensmile]'."
            )  # pragma: no cover
        super().__init__(order=order)
        self.feature_set = feature_set
        self.sample_rate = sample_rate
        self.functionals = functionals
        if self.functionals:
            self.smile = opensmile.Smile(self.feature_set)
        else:
            self.smile = opensmile.Smile(
                self.feature_set,
                feature_level=opensmile.FeatureLevel.LowLevelDescriptors,
            )

    def __call__(self, item: AbstractDataItem) -> AbstractDataItem:
        it = self.smile.process_signal(item.features, self.sample_rate)
        item.features = torch.from_numpy(self.smile.to_numpy(it)).squeeze()
        return item



from concurrent.futures import ThreadPoolExecutor, as_completed
from pathlib import Path
from typing import Dict, List, Optional
from zipfile import ZipFile

import requests
from tqdm import tqdm


class ZipDownloadManager:
    def __init__(
        self,
        files: Dict[str, str],
        path: str,
        max_threads: int = 4,
    ) -> None:
        """Download and extract zip files.

        Args:
            files: Dictionary of filenames and URLs.
            path: Path to download and extract the files to.
            max_threads: Maximum number of threads to use. Defaults to 4.
        """
        self.files = files
        self.path = Path(path)
        self.path.mkdir(parents=True, exist_ok=True)
        self.max_threads = max_threads

    def download(self, check_exist: Optional[List[str]] = None) -> None:
        """Download all files in `files` to `path`.

        Skips files if they are already present or all files in `check_exist`
        are present.

        Args:
            check_exist: List of filenames to check if they already exist.
                Defaults to None.
        """
        if self._check_exist(check_exist):
            print("All files already exist, skipping download.")
            return
        print(f"Downloading {len(self.files)} files to '{self.path}' ...")

        with ThreadPoolExecutor(max_workers=self.max_threads) as executor:
            futures = [
                executor.submit(self._download_file, filename, url)
                for filename, url in self.files.items()
            ]
            for future in as_completed(futures):
                try:
                    future.result()
                except Exception as e:
                    print(f"Error downloading or extracting file: {e}")

        print()

    def extract(self, check_exist: Optional[List[str]] = None) -> None:
        """Extract all files in `files` to `path`.

        Does not extract files if all files in `check_exist` are present.

        Args:
            check_exist: List of filenames to check if they already exist.
                Defaults to None.
        """
        if self._check_exist(check_exist):
            print("All files already exist, skipping extraction.")
            return
        print(f"Extracting {len(self.files)} files to '{self.path}' ...")
        for filename in self.files.keys():
            self._extract_zip(self.path / filename, self.path)

    def _check_exist(self, check_exist: Optional[List[str]] = None) -> bool:
        return check_exist and all(
            (self.path / filename).exists() for filename in check_exist
        )

    def _download_file(self, filename: str, url: str) -> None:
        dest = self.path / filename
        if dest.exists():
            print(f"{dest.name} already exists, skipping download.")
            return
        response = requests.get(url, stream=True)
        total_size = int(response.headers.get("content-length", 0))
        with (
            open(dest, "wb") as file,
            tqdm(
                desc=str(dest.name),
                total=total_size,
                unit="iB",
                unit_scale=True,
                unit_divisor=1024,
            ) as pbar,
        ):
            for data in response.iter_content(chunk_size=1024):
                size = file.write(data)
                pbar.update(size)

    def _extract_zip(self, zip_path: Path, extract_to: Path) -> None:
        if not zip_path.suffix == ".zip":
            return
        with ZipFile(zip_path, "r") as z:
            z.extractall(extract_to)



from typing import Dict, Optional, Tuple, Union

from omegaconf import DictConfig, OmegaConf

import autrainer

from .abstract_transform import AbstractTransform
from .smart_compose import SmartCompose
from .specific_transforms import GrayscaleToRGB, RGBToGrayscale


class TransformManager:
    def __init__(
        self,
        model_transform: Union[DictConfig, Dict],
        dataset_transform: Union[DictConfig, Dict],
        train_augmentation: Optional[SmartCompose] = None,
        dev_augmentation: Optional[SmartCompose] = None,
        test_augmentation: Optional[SmartCompose] = None,
    ) -> None:
        """Transform manager for composing transforms for the train, dev, and
        test datasets.
        Automatically handles the creation of all transformation pipelines
        including incorporating optional augmentations.

        Args:
            model_transform: The model transform configuration.
            dataset_transform: The dataset transform configuration.
            train_augmentation: The train augmentation pipeline.
                Defaults to None.
            dev_augmentation: The dev augmentation pipeline.
                Defaults to None.
            test_augmentation: The test augmentation pipeline.
                Defaults to None.
        """
        if isinstance(model_transform, DictConfig):
            model_transform = OmegaConf.to_container(model_transform)
        if isinstance(dataset_transform, DictConfig):
            dataset_transform = OmegaConf.to_container(dataset_transform)
        self.model_transform = model_transform
        self.dataset_transform = dataset_transform
        self.train_augmentation = train_augmentation
        self.dev_augmentation = dev_augmentation
        self.test_augmentation = test_augmentation

    def get_transforms(
        self,
    ) -> Tuple[SmartCompose, SmartCompose, SmartCompose]:
        """Get the composed transform pipelines for the train, dev, and test
        datasets.

        Returns:
            The composed transform pipelines for the train, dev, and test
                datasets.
        """
        base = self._build("base") + self._match_model_dataset()

        train = base + self._build("train") + self.train_augmentation
        dev = base + self._build("dev") + self.dev_augmentation
        test = base + self._build("test") + self.test_augmentation
        return train, dev, test

    def _build(self, subset: str) -> SmartCompose:
        """Build and combine the model and dataset transforms into a single
        transform. Model transforms override dataset transforms when they share
        the same key (or key + tag).

        Args:
            subset: The subset of the transforms to combine.

        Returns:
            The combined transform.
        """
        combined = []
        seen = set()
        model_transforms = self.model_transform.get(subset, [])
        dataset_transforms = self.dataset_transform.get(subset, [])

        for transform in model_transforms:
            key = self._get_key(transform)
            seen.add(key)

            # allow the model to remove a transform by setting it to None
            # e.g. autrainer.transforms.Normalize: null
            if isinstance(transform, dict) and transform[key] is None:
                continue

            combined.append(transform)

        combined.extend(
            t for t in dataset_transforms if self._get_key(t) not in seen
        )

        return SmartCompose([self._instantiate(t) for t in combined])

    @staticmethod
    def _get_key(transform: Union[str, Dict]) -> str:
        if isinstance(transform, dict):
            return next(iter(transform.keys()))
        return transform

    @staticmethod
    def _instantiate(transform: Union[str, Dict]) -> AbstractTransform:
        if isinstance(transform, str):
            t = transform.split("@")[0]
            return autrainer.instantiate_shorthand(t, AbstractTransform)
        key = next(iter(transform.keys()))
        return autrainer.instantiate_shorthand(
            {key.split("@")[0]: transform[key]},
            AbstractTransform,
        )

    def _match_model_dataset(self) -> SmartCompose:
        """Match the model input type with the dataset type by adding
        the necessary transform.

        Raises:
            ValueError: If the model and dataset types do not match.

        Returns:
            The composed transform to match the model and dataset types.
        """
        m = self.model_transform["type"]
        d = self.dataset_transform["type"]
        match = []
        if m == "image" and d == "grayscale":
            match = [GrayscaleToRGB()]
        elif m == "grayscale" and d == "image":
            match = [RGBToGrayscale()]
        elif m != d:
            raise ValueError(
                f"Model input type '{m}' does not match dataset type '{d}'"
            )
        return SmartCompose(match)



from .abstract_dataset import (
    AbstractDataset,
    BaseClassificationDataset,
    BaseMLClassificationDataset,
    BaseMTRegressionDataset,
    BaseRegressionDataset,
)
from .aibo import AIBO
from .audioset import AudioSet
from .dcase_2016_t1 import DCASE2016Task1
from .dcase_2018_t3 import DCASE2018Task3
from .dcase_2020_t1a import DCASE2020Task1A
from .edansa2019 import EDANSA2019
from .emo_db import EmoDB
from .msppodcast import MSPPodcast
from .speech_commands import SpeechCommands
from .toy_dataset import ToyDataset


__all__ = [
    "AbstractDataset",
    "BaseClassificationDataset",
    "BaseMLClassificationDataset",
    "BaseMTRegressionDataset",
    "BaseRegressionDataset",
    "AIBO",
    "AudioSet",
    "DCASE2016Task1",
    "DCASE2018Task3",
    "DCASE2020Task1A",
    "EDANSA2019",
    "EmoDB",
    "MSPPodcast",
    "SpeechCommands",
    "ToyDataset",
]



from typing import TYPE_CHECKING, Dict

import numpy as np
import torch

from .utils import assert_nonzero_frequency


if TYPE_CHECKING:  # pragma: no cover
    from autrainer.datasets import AbstractDataset


class MSELoss(torch.nn.MSELoss):
    def forward(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        """Wrapper for `torch.nn.MSELoss.forward`.

        Squeezes the model outputs along the last dimension to match the shape
        of the targets.
        Converts the targets to `float`.

        Args:
            x: Batched model outputs.
            y: Targets.

        Returns:
            Loss.
        """
        # TODO: is there a more elegant way to handle broadcasting
        return super().forward(x.squeeze(-1), y.float())


class WeightedMSELoss(MSELoss):
    weights_buffer: torch.Tensor

    def __init__(self, target_weights: Dict[str, float], **kwargs) -> None:
        """Wrapper for `torch.nn.MSELoss` with manual target weights intended
        for multi-target regression tasks.

        The target weights are automatically normalized to sum up to the number
        of targets.

        Args:
            target_weights: Dictionary with target weights corresponding to the
                target labels and their respective weights.
            **kwargs: Additional keyword arguments passed to
                `torch.nn.MSELoss`.
        """
        self.target_weights = target_weights
        super().__init__(**kwargs)

    def setup(self, data: "AbstractDataset") -> None:
        """Calculate the target weights based on the provided dictionary.

        Args:
            data: Instance of the dataset.
        """
        values = []

        if data.task != "mt-regression":
            raise ValueError(
                "`WeightedMSELoss` is only supported for "
                "multi-target regression tasks."
            )

        for target in data.target_transform.target:
            if target not in self.target_weights:
                raise ValueError(
                    f"Missing target weight for target '{target}'."
                )
            values.append(self.target_weights[target])

        assert_nonzero_frequency(np.array(values), len(data.target_transform))
        weight = torch.tensor(values, dtype=torch.float32)
        weight = weight * len(weight) / weight.sum()
        self.register_buffer("weights_buffer", weight)

    def forward(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        """Wrapper for `torch.nn.MSELoss.forward` with manual target weights.

        Args:
            x: Batched model outputs.
            y: Targets.

        Returns:
            Loss.
        """
        return super().forward(x, y) * self.weights_buffer.expand_as(y)



from typing import Callable, List, TypeVar

import numpy as np
import pandas as pd
import torch

from autrainer.core.utils import Bookkeeping
from autrainer.datasets import AbstractDataset


T = TypeVar("T")


class OutputsTracker:
    def __init__(
        self,
        export: bool,
        prefix: str,
        data: AbstractDataset,
        bookkeeping: Bookkeeping,
    ) -> None:
        """Tracker for model outputs, targets, losses, and predictions.

        Args:
            export: Whether to export the results.
            prefix: Prefix for the exported files.
            data: Instance of the dataset.
            bookkeeping: Instance of the bookkeeping class.
        """
        self._export = export
        self._prefix = prefix
        self._data = data
        self._bookkeeping = bookkeeping
        self.reset()

    def reset(self) -> None:
        """Reset the tracker. Clears all the stored data accumulated over a
        single iteration.
        """
        self._outputs = torch.zeros(
            (0, self._data.output_dim), dtype=torch.float32
        )
        self._targets = torch.zeros(
            0, dtype=torch.long
        )  # automatic type promotion in case of float targets
        self._indices = torch.zeros(0, dtype=torch.long)
        self._losses = torch.zeros(0, dtype=torch.float32)
        self._predictions = None
        self._results_df = None

    def update(
        self,
        output: torch.Tensor,
        target: torch.Tensor,
        loss: torch.Tensor,
        sample_idx: torch.Tensor,
    ) -> None:
        """Update the tracker with the current model outputs, targets, losses,
        and sample indices.

        Args:
            output: Detached model outputs.
            target: Targets.
            loss: Per-sample losses.
            sample_idx: Sample indices.
        """
        self._outputs = torch.cat([self._outputs, output.cpu()], dim=0)
        self._targets = torch.cat([self._targets, target.cpu()], dim=0)
        self._losses = torch.cat([self._losses, loss.cpu()], dim=0)
        self._indices = torch.cat([self._indices, sample_idx.cpu()], dim=0)

    def save(self, iteration_folder: str, reset: bool = True) -> None:
        """Save the tracked data to disk.

        Args:
            iteration_folder: Current iteration folder.
            reset: Whether to reset the tracker after saving the results.
                Defaults to True.
        """
        results = {
            "outputs": self._outputs.numpy(),
            "targets": self._targets.numpy(),
            "indices": self._indices.numpy(),
            "losses": self._losses.numpy(),
        }

        _probabilities = self._data.target_transform.probabilities_inference(
            self._outputs
        )
        self._predictions = self._data.target_transform.predict_inference(
            _probabilities
        )
        self._results_df = pd.DataFrame(index=results["indices"])
        self._results_df["predictions"] = self._predictions
        self._results_df["predictions"] = self._results_df[
            "predictions"
        ].apply(self._data.target_transform.decode)
        _probs_df = pd.DataFrame(
            index=results["indices"],
            data=(
                self._data.target_transform.probabilities_to_dict(p)
                for p in _probabilities
            ),
        )
        self._results_df = pd.concat([self._results_df, _probs_df], axis=1)

        if self._export:
            for key, value in results.items():
                self._bookkeeping.save_results_np(
                    value, f"{self._prefix}_{key}.npy", iteration_folder
                )
            self._bookkeeping.save_results_df(
                self._results_df.reset_index(),
                f"{self._prefix}_results.csv",
                iteration_folder,
            )

        if reset:
            self.reset()

    def check_saved(func: Callable[..., T]) -> Callable[..., T]:
        def wrapper(self: "OutputsTracker", *args, **kwargs) -> T:
            if self._results_df is None:
                raise ValueError("Results not saved yet.")
            return func(self, *args, **kwargs)

        return wrapper

    @property
    @check_saved
    def outputs(self) -> np.ndarray:
        """Get the model outputs.

        Returns:
            Model outputs.
        """
        return self._outputs.numpy()

    @property
    @check_saved
    def targets(self) -> np.ndarray:
        """Get the targets.

        Returns:
            Targets.
        """
        return self._targets.numpy()

    @property
    @check_saved
    def indices(self) -> np.ndarray:
        """Get the sample indices.

        Returns:
            Sample indices.
        """
        return self._indices.numpy()

    @property
    @check_saved
    def losses(self) -> np.ndarray:
        """Get the per-sample losses.

        Returns:
            Per-sample losses.
        """
        return self._losses.numpy()

    @property
    @check_saved
    def predictions(self) -> np.ndarray:
        """Get the predictions.

        Returns:
            Predictions.
        """
        return np.array(self._predictions)

    @property
    @check_saved
    def results_df(self) -> pd.DataFrame:
        """Get the results as a DataFrame.

        Returns:
            Results as a DataFrame.
        """
        return self._results_df


def init_trackers(
    exports: List[bool],
    prefixes: List[str],
    data: AbstractDataset,
    bookkeeping: Bookkeeping,
) -> List[OutputsTracker]:
    """Utility function to initialize multiple trackers at once.

    Args:
        exports: Whether to export the results.
        prefixes: Prefixes for the exported files.
        data: Instance of the dataset.
        bookkeeping: Instance of the bookkeeping class.

    Returns:
        List of initialized trackers.
    """
    trackers = []
    for export, prefix in zip(exports, prefixes):
        trackers.append(
            OutputsTracker(
                export=export,
                prefix=prefix,
                data=data,
                bookkeeping=bookkeeping,
            )
        )

    return trackers



import os
import shutil
from typing import Dict, List, Optional, Union

from omegaconf import DictConfig, OmegaConf
import pandas as pd

import autrainer
from autrainer.core.constants import NamingConstants
from autrainer.core.plotting import PlotMetrics
from autrainer.loggers import get_params_to_export
from autrainer.metrics import AbstractMetric

from .postprocessing_utils import (
    get_plotting_params,
    get_run_names,
    get_training_type,
    load_yaml,
    save_yaml,
)


class SummarizeGrid:
    def __init__(
        self,
        results_dir: str,
        experiment_id: str,
        summary_dir: str = "summary",
        training_dir: str = "training",
        clear_old_outputs: bool = True,
        training_type: Optional[str] = None,
        max_runs_plot: Optional[int] = None,
        plot_params: Optional[Union[DictConfig, Dict]] = None,
    ) -> None:
        """Summarize the results of a grid search.

        Args:
            results_dir: The directory where the results are stored.
            experiment_id: The ID of the grid search experiment.
            summary_dir: The directory where the the grid search summary will
                be stored. Defaults to "summary".
            training_dir: The directory of the training results of the
                    experiment. Defaults to "training".
            clear_old_outputs: Whether to clear existing summary outputs.
                Defaults to True.
            training_type: The type of training in ["Epoch", "Step"]. If None,
                it will be inferred from the training results. Defaults to None.
            max_runs_plot: The maximum number of best runs to plot. If None,
                all runs will be plotted. Defaults to None.
            plot_params: Additional parameters for plotting. Defaults to None.
        """
        self.results_dir = results_dir
        self.experiment_id = experiment_id
        self.output_directory = os.path.join(
            self.results_dir, self.experiment_id, summary_dir
        )
        self.training_directory = os.path.join(
            self.results_dir, self.experiment_id, training_dir
        )
        if os.path.exists(self.output_directory) and clear_old_outputs:
            shutil.rmtree(self.output_directory)
        os.makedirs(self.output_directory, exist_ok=True)

        self.run_names = get_run_names(self.training_directory)
        self.training_type = training_type
        if self.training_type is None:
            self.training_type = get_training_type(
                self.training_directory, self.run_names
            )
        self.max_runs_plot = max_runs_plot
        self.plot_params = plot_params
        self.run_names.sort()
        if self.plot_params is None:
            self.plot_params = get_plotting_params(
                self.training_directory, self.run_names[0]
            )
        self.metric_fns: List[AbstractMetric] = []

    def summarize(self) -> None:
        """Summarize the results of the grid search."""
        self._summarize_metrics()
        self._summarize_config()
        self._summarize_best_runs()

    def _summarize_metrics(self) -> None:
        records = []
        for n in self.run_names:
            metrics = self._read_metrics(n)
            test = self._read_test_metrics(n)
            records.append({"run_name": n, **metrics, **test})
        df = pd.DataFrame.from_records(records)
        path = os.path.join(self.output_directory, "metrics.csv")
        df.to_csv(path, index=False)

    def _summarize_config(self) -> None:
        records = []
        for n in self.run_names:
            run_dir = os.path.join(self.training_directory, n)
            config_path = os.path.join(run_dir, ".hydra", "config.yaml")
            config = load_yaml(config_path)
            config = dict(sorted(get_params_to_export(config).items()))
            time = self._read_times(n)
            records.append({"run_name": n, **config, **time})
        df = pd.DataFrame.from_records(records)
        path = os.path.join(self.output_directory, "config.csv")
        df.to_csv(path, index=False)

    def _summarize_best_runs(self) -> None:
        summary_path = os.path.join(self.output_directory, "metrics.csv")
        df = pd.read_csv(summary_path)
        metric_cols = [m for m in df.columns if not m.endswith(".std")]
        metric_cols.remove("run_name")
        metric_cols.remove("best_iteration")
        best_runs = {}
        for metric in metric_cols:
            if "loss" in metric:
                idx = df[metric].idxmin()
            else:
                base_metric = metric.replace("test_", "").replace("train_", "")
                m = next(
                    (fn for fn in self.metric_fns if fn.name == base_metric),
                    None,
                )
                if m is None:
                    raise ValueError(f"No metric found for '{metric}'.")
                idx = m.get_best_pos(df[metric])
            best_runs[metric] = df.loc[idx, "run_name"]
        path = os.path.join(self.output_directory, "best_runs.yaml")
        save_yaml(path, best_runs)

    def _find_metrics_to_plot(self) -> List[str]:
        summary_path = os.path.join(self.output_directory, "metrics.csv")
        df = pd.read_csv(summary_path)
        plot_metrics = [
            m
            for m in list(df.columns)
            if m in [fn.name for fn in self.metric_fns]
            or m in ["train_loss", "dev_loss"]
        ]
        return plot_metrics

    def plot_metrics(self) -> None:
        """Plot the metrics of the grid search."""
        plotter = PlotMetrics(
            output_directory=self.output_directory,
            training_type=self.training_type,
            metric_fns=self.metric_fns,
            **self.plot_params,
        )
        metrics_to_plot = self._find_metrics_to_plot()
        for metric in metrics_to_plot:
            metrics_list = []
            metrics_std_list = []
            for n in self.run_names:
                metrics_path = os.path.join(
                    self.training_directory, n, "metrics.csv"
                )
                df = pd.read_csv(metrics_path, index_col="iteration")
                if metric not in df.columns:
                    continue
                metrics_list.append(df[[metric]].rename(columns={metric: n}))

                std_metric = f"{metric}.std"
                if std_metric in df.columns:
                    metrics_std_list.append(
                        df[[std_metric]].rename(columns={std_metric: n})
                    )

            metrics_df = pd.concat(metrics_list, axis=1)
            metrics_std_df = (
                pd.concat(metrics_std_list, axis=1)
                if metrics_std_list
                else None
            )
            plotter.plot_metric(
                metrics_df, metric, metrics_std_df, max_runs=self.max_runs_plot
            )

    def plot_aggregated_bars(self) -> None:
        """Plot additional aggregated bars for the grid search."""
        metrics_to_plot = self._find_metrics_to_plot()
        plotter = PlotMetrics(
            output_directory=self.output_directory,
            training_type=self.training_type,
            metric_fns=self.metric_fns,
            **self.plot_params,
        )
        metrics_df = pd.DataFrame()

        # Iterate over each CSV file and build up metrics dataframe
        for n in self.run_names:
            metrics_path = os.path.join(
                self.training_directory, n, "metrics.csv"
            )
            df = pd.read_csv(metrics_path)
            last_row = df.iloc[-1:]
            run_details = n.split("_")
            for i, param in enumerate(NamingConstants().NAMING_CONVENTION):
                last_row.insert(i, param, run_details[i])
            metrics_df = pd.concat((metrics_df, last_row), ignore_index=True)
        for metric in metrics_to_plot:
            plotter.plot_aggregated_bars(metrics_df, metric)

    def _read_metrics(self, run_name: str) -> dict:
        metrics_path = os.path.join(
            self.training_directory, run_name, "metrics.csv"
        )
        config_path = os.path.join(
            self.training_directory, run_name, ".hydra", "config.yaml"
        )
        df = pd.read_csv(metrics_path, index_col="iteration")
        cfg = OmegaConf.load(config_path)
        tracking_metric = autrainer.instantiate_shorthand(
            config=cfg.dataset.tracking_metric,
            instance_of=AbstractMetric,
        )
        self._update_metric_fns(cfg.dataset.metrics)
        best_iteration = tracking_metric.get_best_pos(df[tracking_metric.name])
        metrics_dict = df.loc[best_iteration].to_dict()
        metrics_dict["best_iteration"] = best_iteration
        return metrics_dict

    def _read_test_metrics(self, run_name: str, prefix: str = "test_") -> dict:
        path = os.path.join(
            self.training_directory, run_name, "_test", "test_holistic.yaml"
        )
        test_dict = load_yaml(path)
        test_dict = {prefix + k: v["all"] for k, v in test_dict.items()}
        return test_dict

    def _read_times(self, run_name: str) -> dict:
        path = os.path.join(self.training_directory, run_name, "timer.yaml")
        time_dict = load_yaml(path)
        return {
            "time_train_mean": time_dict["train"]["mean"],
            "time_dev_mean": time_dict["dev"]["mean"],
            "time_test_mean": time_dict["test"]["mean"],
        }

    def _update_metric_fns(self, metric_cfg: Union[dict, DictConfig]) -> None:
        metrics = [
            autrainer.instantiate_shorthand(m, instance_of=AbstractMetric)
            for m in metric_cfg
        ]
        for m in metrics:
            if m.name in [m.name for m in self.metric_fns]:
                continue
            self.metric_fns.append(m)



import os
from typing import List

from omegaconf import DictConfig
import yaml


def get_run_names(training_directory: str):
    run_names = []
    for dir in os.listdir(training_directory):
        if os.path.isdir(os.path.join(training_directory, dir)):
            run_names.append(dir)
    return run_names


def load_yaml(path: str):
    with open(path, "r") as f:
        return yaml.safe_load(f)


def save_yaml(path: str, data: dict):
    with open(path, "w") as f:
        yaml.safe_dump(data, f)


def get_training_type(training_directory: str, runs: List[str]):
    training_types = []
    for run in runs:
        path = os.path.join(training_directory, run, ".hydra", "config.yaml")
        config = DictConfig(load_yaml(path))
        training_types.append(config.training_type)
    assert (
        len(set(training_types)) == 1
    ), f"Multiple training types found: {set(training_types)}"
    return training_types[0]


def get_plotting_params(training_directory: str, run: str):
    path = os.path.join(training_directory, run, ".hydra", "config.yaml")
    return DictConfig(load_yaml(path)).plotting
